#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import abc
import ctypes
import os
from datetime import datetime
from typing import List, Union, Optional, Any

import ROOT
import numpy

import wagascianpy.analysis.spill as spill
import wagascianpy.database.bsddb as bsddb
import wagascianpy.database.wagascidb as wagascidb
import wagascianpy.plotting as pl
import wagascianpy.plotting.colors as colors
import wagascianpy.plotting.graph as graph
import wagascianpy.plotting.harvest as harvest
import wagascianpy.plotting.marker as marker
import wagascianpy.plotting.topology as topol

ROOT.PyConfig.IgnoreCommandLineOptions = True

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3


class Plotter(ABC):

    def __init__(self,
                 output_file_path="./plot.pdf",
                 save_tfile=False,
                 markers=marker.MarkerTuple(run=False, maintenance=False, trouble=False, bsd=False)):
        # type: (str, bool, marker.MarkerTuple) -> None
        ROOT.PyConfig.IgnoreCommandLineOptions = True
        ROOT.gROOT.IsBatch()
        self._canvas = ROOT.TCanvas("canvas", "canvas", 1280, 720)
        self._title = ""
        self._graphs = []
        self._markers = []
        self._multi_graph = None
        self._1d_draw_options = "AL"
        self._2d_draw_options = "ZCOLPCOL"
        self._output_file_path = output_file_path
        self._enabled_markers = markers
        self._save_tfile = save_tfile
        self._plot_legend_flag = False
        self._xdata_is_datetime = True
        self._yrange = (None, None)
        self._logscale = False

    def template_plotter(self):
        # type: (...) -> None
        self._graphs = self.setup_graphs()
        self.set_title()
        for gr in self._graphs:
            self.gather_data(gr)
        self.change_graph_titles(self._graphs)
        self.build_multigraph()
        self.change_yaxis_title(self._multi_graph)
        self.add_run_markers()
        self.add_maintenance_day_markers()
        self.add_trouble_markers()
        self.add_bsd_markers()
        self.plot()
        if self._save_tfile:
            self.save()
        self._canvas.Close()

    @property
    def plot_legend_flag(self):
        # type: (...) -> bool
        return self._plot_legend_flag

    @plot_legend_flag.setter
    def plot_legend_flag(self, plot_legend_flag):
        # type: (bool) -> None
        if not isinstance(plot_legend_flag, bool):
            raise TypeError("Plot legend flag must be a boolean")
        self._plot_legend_flag = plot_legend_flag

    @property
    def draw_options_1d(self):
        # type: (...) -> str
        return self._1d_draw_options

    @draw_options_1d.setter
    def draw_options_1d(self, draw_options):
        # type: (str) -> None
        self._1d_draw_options = draw_options

    @property
    def draw_options_2d(self):
        # type: (...) -> str
        return self._2d_draw_options

    @draw_options_2d.setter
    def draw_options_2d(self, draw_options):
        # type: (str) -> None
        self._2d_draw_options = draw_options

    @property
    def xdata_is_datetime(self):
        # type: (...) -> bool
        return self._xdata_is_datetime

    @xdata_is_datetime.setter
    def xdata_is_datetime(self, xdata_is_datetime):
        # type: (bool) -> None
        assert isinstance(xdata_is_datetime, bool), "xdata_is_datetime only accept a boolean value"
        self._xdata_is_datetime = xdata_is_datetime

    @property
    def logscale(self):
        # type: (...) -> bool
        return self._logscale

    @logscale.setter
    def logscale(self, logscale):
        # type: (bool) -> None
        self._logscale = logscale

    @abc.abstractmethod
    def set_title(self):
        # type: (...) -> None
        if self._topology.how_many_enabled() == 1:
            self._title += " for {}".format(self._topology.get_enabled()[0].name)
        elif self._topology.how_many_enabled() == 2:
            self._title += " for {} and {}".format(self._topology.get_enabled()[0].name,
                                                   self._topology.get_enabled()[1].name)

    @abc.abstractmethod
    def setup_graphs(self):
        # type: (...) -> None
        pass

    @abc.abstractmethod
    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        pass

    def build_multigraph(self):
        # type: (...) -> None
        self._multi_graph = ROOT.TMultiGraph()
        self._multi_graph.SetName("multi_graph")
        ROOT.TGaxis.SetMaxDigits(3)
        for gr in self._graphs:
            if not gr.is_empty() and not isinstance(gr, graph.Graph2D):
                self._multi_graph.Add(gr.make_tgraph())
                if gr.yaxis_color != colors.Colors.Black.value:
                    self._multi_graph.GetYaxis().SetTitleColor(gr.yaxis_color)
                    self._multi_graph.GetYaxis().SetLabelColor(gr.yaxis_color)
                    self._multi_graph.GetYaxis().SetAxisColor(gr.yaxis_color)
        self._multi_graph.SetTitle(self._title)
        if self._xdata_is_datetime:
            self._multi_graph.GetXaxis().SetNdivisions(9, 3, 0, ROOT.kTRUE)
            self._multi_graph.GetXaxis().SetTimeDisplay(1)
            self._multi_graph.GetXaxis().SetTimeFormat("#splitline{%d %b}{%H:%M}%F1970-01-01 00:00:00")
            self._multi_graph.GetXaxis().SetLabelOffset(0.03)
        if self._yrange[0] is not None and self._yrange[1] is not None:
            self._multi_graph.GetYaxis().SetRangeUser(self._yrange[0], self._yrange[1])

    def add_maintenance_day_markers(self):
        # type: (...) -> None
        if self._enabled_markers.maintenance and hasattr(self, "_start") and hasattr(self, "_stop"):
            self.make_maintenance_day_markers(start=self._start, stop=self._stop,
                                              wagasci_database=self._patron.wagasci_database)

    def add_run_markers(self):
        # type: (...) -> None
        if self._enabled_markers.run and hasattr(self, "_start") and hasattr(self, "_stop"):
            self.make_run_markers(start=self._start, stop=self._stop,
                                  wagasci_database=self._patron.wagasci_database)

    def add_bsd_markers(self):
        # type: (...) -> None
        if self._enabled_markers.bsd and hasattr(self, "_start") and hasattr(self, "_stop"):
            self.make_bsd_markers(start=self._start, stop=self._stop,
                                  bsd_database=self._bsd_database if hasattr(self, "_bsd_database") else None)

    def add_trouble_markers(self):
        # type: (...) -> None
        if self._enabled_markers.trouble and hasattr(self, "_start") and hasattr(self, "_stop"):
            self.make_trouble_markers(start=self._start, stop=self._stop,
                                      wagasci_database=self._patron.wagasci_database)

    def change_graph_titles(self, graphs):
        # type: (List[graph.Graph]) -> None
        pass

    def change_yaxis_title(self, multigraph):
        # type: (ROOT.TMultiGraph) -> None
        pass

    def make_run_markers(self, wagasci_database, start, stop=None):
        # type: (str, Union[int, datetime], Optional[Union[int, datetime]]) -> None
        with wagascidb.WagasciDataBase(db_location=wagasci_database) as db:
            if isinstance(start, IntTypes):
                if not stop:
                    stop = db.get_last_run_number(only_good=False)
                records = db.get_run_interval(run_number_start=start, run_number_stop=stop, only_good=False)
            else:
                if not stop:
                    stop = datetime.now()
                records = db.get_time_interval(datetime_start=start, datetime_stop=stop, only_good=False,
                                               include_overlapping=False)
        counter = 0
        markers = []
        for record in records:
            mk = marker.DoubleMarker(left_position=record["start_time"],
                                     right_position=record["stop_time"],
                                     left_text="WAGASCI run %s" % record["run_number"],
                                     right_text="",
                                     line_color=colors.Colors.Blue)
            if counter % 2 == 0:
                mk.fill_color = colors.Colors.Azure.value
            else:
                mk.fill_color = colors.Colors.Orange.value
            mk.transparency = 0.1
            markers.append(mk)
            counter += 1
        self._markers += markers

    def make_bsd_markers(self, bsd_database, start, stop=None):
        # type: (str, datetime, Optional[datetime]) -> None
        with bsddb.BsdDataBase(bsd_database_location=bsd_database) as db:
            if not stop:
                stop = datetime.now()
            records = db.get_time_interval(datetime_start=start, datetime_stop=stop, only_good=False,
                                           include_overlapping=False)
        counter = 0
        markers = []
        for record in records:
            mk = marker.DoubleMarker(left_position=record["start_time"],
                                     right_position=record["stop_time"],
                                     left_text=record["name"],
                                     right_text="",
                                     line_color=colors.Colors.Blue)
            if counter % 2 == 0:
                mk.fill_color = colors.Colors.Azure.value
            else:
                mk.fill_color = colors.Colors.Orange.value
            mk.transparency = 0.1
            markers.append(mk)
            counter += 1
        self._markers += markers

    def make_maintenance_day_markers(self, wagasci_database, start, stop=None):
        # type: (str, Union[int, datetime], Optional[Union[int, datetime]]) -> None
        markers = marker.MaintenanceDays(
            start=start, stop=stop, wagasci_database=wagasci_database
        ).get_markers(include_overlapping=False)
        self._markers += markers

    def make_trouble_markers(self, wagasci_database, start, stop=None):
        # type: (str, Union[int, datetime], Optional[Union[int, datetime]]) -> None
        markers = marker.TroubleEvents(
            start=start, stop=stop, wagasci_database=wagasci_database
        ).get_markers(include_overlapping=False)
        self._markers += markers

    def _has_both_graph1d_and_graph2d(self):
        # type: (...) -> bool
        has_graph1d = self._how_many_graph1d() > 0
        has_graph2d = self._how_many_graph2d() > 0
        return has_graph2d and has_graph1d

    def _how_many_graph1d(self):
        # type: (...) -> int
        return sum(map(lambda g: not isinstance(g, graph.Graph2D) and not g.is_empty(),
                       self._graphs))

    def _how_many_graph2d(self):
        # type: (...) -> int
        return sum(map(lambda g: isinstance(g, graph.Graph2D) and not g.is_empty(),
                       self._graphs))

    def plot(self):
        # type: (...) -> None

        # PREPARE CANVAS
        ROOT.gStyle.Reset("Modern")
        ROOT.gStyle.SetTitleFontSize(0.035)
        pad1 = ROOT.TPad("pad1", "", 0, 0, 1, 1)
        pad2 = ROOT.TPad("pad2", "", 0, 0, 1, 1)
        self._canvas.cd()
        if self._has_both_graph1d_and_graph2d():
            self._canvas.SetFillColor(0)
            self._canvas.SetBorderMode(0)
            pad1.SetGrid()
            pad2.SetFillStyle(4000)
            pad2.SetFrameFillStyle(0)
            pad1.Draw()
            pad1.cd()

        # PLOT 2D HISTOGRAMS
        for gr in self._graphs:
            if isinstance(gr, pl.graph.Graph2D) and not gr.is_empty():
                tgraph = gr.make_tgraph(self.logscale)
                tgraph.SetTitle(self._title)
                tgraph.Draw(self.draw_options_2d)
                tgraph.GetYaxis().SetTitleOffset(0.75)
                if self.logscale:
                    self._canvas.SetLogy()
                    pad1.SetLogy()

        # PLOT 1D GRAPHS
        ROOT.gPad.Update()
        xmin = ctypes.c_double(0)
        ymin = ctypes.c_double(0)
        xmax = ctypes.c_double(0)
        ymax = ctypes.c_double(0)
        if self._has_both_graph1d_and_graph2d():
            pad1.GetRangeAxis(xmin, ymin, xmax, ymax)
            ymin = self._multi_graph.GetYaxis().GetXmin()
            ymax = self._multi_graph.GetYaxis().GetXmax()
            pad2.RangeAxis(xmin.value, ymin, xmax.value, ymax)
            pad2.Draw()
            pad2.cd()
        if self._how_many_graph1d() > 0:
            self._multi_graph.Draw(self.draw_options_1d)

        # PLOT MARKERS
        ROOT.gPad.Update()
        tobjects = []
        for mk in self._markers:
            tobjects += mk.make_tobjects()
        for tobj in tobjects:
            tobj.Draw()
        if self._plot_legend_flag:
            tlegend = ROOT.TLegend(0.13, 0.7, 0.4, 0.89)
            tlegend.SetFillColorAlpha(ROOT.kWhite, 1.)
            for gr in [gr for gr in self._graphs if not gr.is_empty()]:
                opt = "f"
                if "l" in self.draw_options_1d.lower():
                    opt += "l"
                if "p" in self.draw_options_1d.lower():
                    opt += "p"
                tlegend.AddEntry(gr.id, gr.title, opt)
            tlegend.Draw()

        # PRINT TO FILE
        ROOT.gPad.Update()
        self._canvas.Print(self._output_file_path)

    def save(self):
        # type: (...) -> None
        output_path = os.path.splitext(self._output_file_path)[0] + ".root"
        output_tfile = ROOT.TFile(output_path, "RECREATE")
        output_tfile.cd()
        self._canvas.Write()
        self._multi_graph.Write()
        output_tfile.Write()
        output_tfile.Close()


class BsdPlotter(Plotter, ABC):

    def __init__(self,
                 bsd_database,  # type: str
                 bsd_repository,  # type: str
                 start,  # type: Union[int, str, datetime]
                 stop=None,  # type: Optional[int, str, datetime]
                 wagasci_database=None,  # type: Optional[str]
                 t2krun=10,  # type: int
                 only_good=False,  # type: bool
                 *args, **kwargs):
        super(BsdPlotter, self).__init__(*args, **kwargs)
        self._bsd_database = bsd_database
        self._bsd_repository = bsd_repository
        self._t2krun = t2krun
        self._only_good = only_good
        self._start = start
        self._stop = stop
        self._patron = harvest.Patron(start=start, stop=stop, wagasci_database=wagasci_database)
        self._bsd_harvester_class = None

    @abc.abstractmethod
    def set_title(self):
        # type: (...) -> None
        pass

    @abc.abstractmethod
    def setup_graphs(self):
        # type: (...) -> None
        pass

    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        assert self._bsd_harvester_class is not None, "Derived class must set the _bsd_harvester_class attribute"
        if gr.id != "BSD":
            raise ValueError("Wrong graph with title {} and ID {}".format(gr.title, gr.id))
        self._patron.harvester = self._bsd_harvester_class(database=self._bsd_database, repository=self._bsd_repository,
                                                           t2krun=self._t2krun)
        gr.xdata, gr.ydata = self._patron.gather_data(only_good=self._only_good)


class BsdPotPlotter(BsdPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(BsdPotPlotter, self).__init__(*args, **kwargs)
        self._bsd_harvester_class = harvest.BsdPotHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "Delivered POT during run {};;POT".format(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        bsd_pot = graph.Graph("Delivered POT", "BSD")
        bsd_pot.color = colors.Colors.Red.value
        return [bsd_pot]


class BsdSpillPlotter(BsdPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(BsdSpillPlotter, self).__init__(*args, **kwargs)
        self._bsd_harvester_class = harvest.BsdSpillHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "BSD spill history during run {};;spill number".format(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        bsd = graph.Graph("BSD spill history", "BSD")
        bsd.color = colors.Colors.Red.value
        return [bsd]


class WagasciPlotter(Plotter, ABC):

    def __init__(self,
                 bsd_database,  # type: str
                 bsd_repository,  # type: str
                 wagasci_database,  # type: str
                 wagasci_repository,  # type: str
                 start,  # type: Union[str, int]
                 stop=None,  # type: Optional[Union[str, int]]
                 topology=None,  # type: Optional[pl.topology.Topology]
                 t2krun=10,  # type: int
                 only_good=False,  # type: bool
                 *args, **kwargs):
        # type: (...) -> None
        super(WagasciPlotter, self).__init__(*args, **kwargs)
        self._bsd_database = bsd_database
        self._bsd_repository = bsd_repository
        self._wagasci_repository = wagasci_repository
        self._t2krun = t2krun
        self._start = start
        self._stop = stop
        self._patron = harvest.Patron(start=start, stop=stop, wagasci_database=wagasci_database)
        self._wagasci_harvester_class = None
        self._bsd_harvester_class = None
        self._topology = topology if topology is not None else pl.topology.Topology()
        self._only_good = only_good

    @abc.abstractmethod
    def set_title(self):
        # type: (...) -> None
        super(WagasciPlotter, self).set_title()

    @abc.abstractmethod
    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        graphs = []
        for enabled_detector in self._topology.get_enabled():
            gr = graph.Graph(enabled_detector.name)
            gr.color = enabled_detector.name
            graphs.append(gr)
        return graphs

    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        if gr.id == "BSD":
            assert self._bsd_harvester_class is not None, \
                "Derived class must set the _bsd_harvester_class attribute"
            if not self._patron.is_harvester_ready() or \
                    not isinstance(self._patron.harvester, self._bsd_harvester_class):
                self._patron.harvester = self._bsd_harvester_class(
                    database=self._bsd_database,
                    repository=self._bsd_repository,
                    t2krun=self._t2krun)
            gr.xdata, gr.ydata = self._patron.gather_data(only_good=self._only_good)
        else:
            for enabled_detector in self._topology.get_enabled():
                if gr.id == str(enabled_detector.name):
                    assert self._wagasci_harvester_class is not None, \
                        "Derived class must set the _wagasci_harvester_class attribute"
                    if not self._patron.is_harvester_ready() or \
                            not isinstance(self._patron.harvester, self._wagasci_harvester_class):
                        self._patron.harvester = self._wagasci_harvester_class(
                            database=self._patron.wagasci_database,
                            repository=self._wagasci_repository,
                            t2krun=self._t2krun,
                            topology=self._topology)
                    gr.xdata, gr.ydata = self._patron.gather_data(detector_name=enabled_detector.name,
                                                                  only_good=self._only_good)


class WagasciPotPlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(WagasciPotPlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.WagasciPotHarvester
        self._bsd_harvester_class = harvest.BsdPotHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "#splitline{Accumulated POT for each subdetector during run %s}" \
                      "{after spill matching but before ADC calibration};;POT" % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = True
        self.xdata_is_datetime = True
        graphs = super(WagasciPotPlotter, self).setup_graphs()
        bsd_graph = graph.Graph("Delivered POT", "BSD")
        bsd_graph.color = colors.Colors.Red.value
        graphs.append(bsd_graph)
        return graphs

    # noinspection PyUnresolvedReferences
    def change_graph_titles(self, graphs):
        # type: (List[pl.graph.Graph]) -> None
        bsd_graph = next((bsd_graph for bsd_graph in graphs if bsd_graph.id == "BSD"), None)
        if bsd_graph is None:
            return
        if bsd_graph.ydata.size == 0:
            bsd_pot = 0
        else:
            bsd_pot = numpy.amax(bsd_graph.ydata)
        bsd_graph.title += " = {:.2e} POT".format(bsd_pot)
        for igraph in [gr for gr in graphs if gr.id != "BSD"]:
            if igraph.ydata.size == 0:
                max_pot = 0
            else:
                max_pot = numpy.amax(igraph.ydata)
            percent = 100 * float(max_pot) / float(bsd_pot) if bsd_pot != 0 else 0
            igraph.title += " {:.1f}%".format(percent)


class WagasciSpillHistoryPlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(WagasciSpillHistoryPlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.WagasciSpillHistoryHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "#splitline{WAGASCI spill history during run %s}" \
                      "{before bit flip fixing};;spill number" % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_1d = "AP"
        graphs = super(WagasciSpillHistoryPlotter, self).setup_graphs()
        for gr in graphs:
            gr.yrange = graph.Range(lower_bound=spill.WAGASCI_MINIMUM_SPILL, upper_bound=spill.WAGASCI_MAXIMUM_SPILL)
        return graphs


class WagasciFixedSpillPlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(WagasciFixedSpillPlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.WagasciFixedSpillHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "#splitline{WAGASCI fixed spill history during run %s}" \
                      "{after bit flip fixing but before ADC calibration};;spill number" % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_1d = "AP"
        return super(WagasciFixedSpillPlotter, self).setup_graphs()


class WagasciSpillNumberPlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(WagasciSpillNumberPlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.WagasciSpillNumberHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "#splitline{WAGASCI spill number during run %s}" \
                      "{before bit flip fixing and BSD spill matching};event number (increasing in time);spill number" \
                      % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = False
        self.draw_options_1d = "AP"
        graphs = super(WagasciSpillNumberPlotter, self).setup_graphs()
        for gr in graphs:
            gr.yrange = graph.Range(lower_bound=spill.WAGASCI_MINIMUM_SPILL, upper_bound=spill.WAGASCI_MAXIMUM_SPILL)
        return graphs


class TemperaturePlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(TemperaturePlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.TemperatureHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "WAGASCI temperature history during run %s};;Temperature (C°)" % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = True
        self.xdata_is_datetime = True
        return super(TemperaturePlotter, self).setup_graphs()


class HumidityPlotter(WagasciPlotter):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(HumidityPlotter, self).__init__(*args, **kwargs)
        self._wagasci_harvester_class = harvest.HumidityHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "WAGASCI humidity history during run %s};;Humidity (%%)" % str(self._t2krun)

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = True
        self.xdata_is_datetime = True
        return super(HumidityPlotter, self).setup_graphs()


class AdcCalibrationPlotterDif(Plotter, ABC):

    def __init__(self,
                 start,  # type: Union[str, int]
                 stop=None,  # type: Optional[Union[str, int]]
                 wagasci_database=None,  # type: Optional[str]
                 history_location=None,  # type: Optional[str]
                 topology=None,  # type: Optional[pl.topology.Topology]
                 *args, **kwargs):
        # type: (...) -> None
        super(AdcCalibrationPlotterDif, self).__init__(*args, **kwargs)
        self._start = start
        self._stop = stop
        self._history_location = history_location
        self._topology = topology
        self._data_harvester_class = None
        self._temperature_harvester_class = None
        self._patron = harvest.Patron(start=start, stop=stop, wagasci_database=wagasci_database)

    @abc.abstractmethod
    def set_title(self):
        # type: (...) -> None
        super(AdcCalibrationPlotterDif, self).set_title()

    @abc.abstractmethod
    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        graphs = [graph.Graph2D(title="history", graph_id="data")]
        if self._topology.how_many_enabled() == 1:
            for enabled_detector in self._topology.get_enabled():
                gr = graph.Graph(title=enabled_detector.name, graph_id="temperature")
                gr.color = colors.Colors.Red
                gr.yaxis_color = colors.Colors.Red
                graphs.append(gr)
        return graphs

    def change_yaxis_title(self, multigraph):
        # type: (ROOT.TMultiGraph) -> None
        multigraph.GetYaxis().SetTitle("Temperature (Celsius Degrees)")

    def _set_harvester(self, harvester_class):
        assert harvester_class is not None, "Derived class must set the harvester class attribute"
        if not self._patron.is_harvester_ready() or not isinstance(self._patron.harvester, harvester_class):
            self._patron.harvester = harvester_class(database=self._patron.wagasci_database,
                                                     repository=self._history_location,
                                                     t2krun=None,
                                                     topology=self._topology)

    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        if not self._topology.iterate_by_dif:
            raise RuntimeError("You must iterate through the active detectors by DIF")
        if gr.id == "data":
            self._set_harvester(self._data_harvester_class)
            gr.xdata, gr.ydata = self._patron.gather_data()
        else:
            for enabled_detector in self._topology.get_enabled():
                if gr.id == str(enabled_detector.name):
                    self._set_harvester(self._temperature_harvester_class)
                    gr.xdata, gr.ydata = self._patron.gather_data(detector_name=enabled_detector.name)


class GainHistoryPlotterDif(AdcCalibrationPlotterDif):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(GainHistoryPlotterDif, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.GainHarvesterDif
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "Gain history"
        super(GainHistoryPlotterDif, self).set_title()
        self._title += ";;Gain (ADC counts)"

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_2d = "COL"
        if self._topology.how_many_enabled() == 1:
            self.draw_options_2d += "Y+"
        else:
            self.draw_options_2d += "Z"
        self.draw_options_1d = "AL"
        return super(GainHistoryPlotterDif, self).setup_graphs()


class DarkNoiseHistoryPlotterDif(AdcCalibrationPlotterDif):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(DarkNoiseHistoryPlotterDif, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.DarkNoiseHarvesterDif
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "Dark noise history"
        super(DarkNoiseHistoryPlotterDif, self).set_title()
        self._title += ";;Dark noise (Hz)"

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_2d = "COL"

        if self._topology.how_many_enabled() == 1:
            self.draw_options_2d += "Y+"
        else:
            self.draw_options_2d += "Z"
        self.draw_options_1d = "AL"
        self.logscale = True

        return super(DarkNoiseHistoryPlotterDif, self).setup_graphs()


class ThresholdHistoryPlotterDif(AdcCalibrationPlotterDif):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(ThresholdHistoryPlotterDif, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.ThresholdHarvesterDif
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester
        self._yrange = (1, 6)

    def set_title(self):
        # type: (...) -> None
        self._title = "Hit threshold level history"
        super(ThresholdHistoryPlotterDif, self).set_title()
        self._title += ";;Hit threshold (PEU)"

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_2d = "COL"

        if self._topology.how_many_enabled() == 1:
            self.draw_options_2d += "Y+"
        else:
            self.draw_options_2d += "Z"
        self.draw_options_1d = "AL"
        self.logscale = False

        return super(ThresholdHistoryPlotterDif, self).setup_graphs()


class AdcCalibrationPlotterChannel(Plotter, ABC):

    def __init__(self,
                 start,  # type: Union[str, int]
                 stop=None,  # type: Optional[Union[str, int]]
                 wagasci_database=None,  # type: Optional[str]
                 history_location=None,  # type: Optional[str]
                 dif=0,  # type: int
                 chip=0,  # type: int
                 channel=0,  # type: int
                 *args, **kwargs):
        super(AdcCalibrationPlotterChannel, self).__init__(*args, **kwargs)
        self._data_harvester_class = None
        self._temperature_harvester_class = None
        self._start = start
        self._stop = stop
        self._history_location = history_location
        self._dif = dif
        self._chip = chip
        self._channel = channel
        self._topology = topol.Topology(iterate_by_dif=True)
        self._topology.disable_all_but(topol.DifIndex.get_name(index=dif))
        self._patron = harvest.Patron(start=start, stop=stop, wagasci_database=wagasci_database)

    @abc.abstractmethod
    def set_title(self):
        # type: (...) -> None
        pass

    def _make_data_graph_title(self):
        # type: (...) -> str
        return "{} : CHIP {} CHANNEL {}".format(pl.topology.DifIndex.get_name(self._dif), self._chip, self._channel)

    @abc.abstractmethod
    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        graphs = [graph.Graph(title=self._make_data_graph_title(), graph_id="data")]
        gr = graph.Graph(title=pl.topology.DifIndex.get_name(self._dif), graph_id="temperature")
        gr.color = colors.Colors.Red
        gr.yaxis_color = colors.Colors.Red
        graphs.append(gr)
        return graphs

    # def change_yaxis_title(self, multigraph):
    #     # type: (ROOT.TMultiGraph) -> None
    #     multigraph.GetYaxis().SetTitle("Temperature (Celsius Degrees)")

    def _set_data_harvester(self, harvester_class):
        assert harvester_class is not None, "Derived class must set the harvester class attribute"
        if not self._patron.is_harvester_ready() or not isinstance(self._patron.harvester, harvester_class):
            self._patron.harvester = harvester_class(database=self._patron.wagasci_database,
                                                     repository=self._history_location,
                                                     t2krun=None,
                                                     topology=self._topology,
                                                     chip=self._chip,
                                                     channel=self._channel)

    def _set_temp_harvester(self, harvester_class):
        assert harvester_class is not None, "Derived class must set the harvester class attribute"
        if not self._patron.is_harvester_ready() or not isinstance(self._patron.harvester, harvester_class):
            self._patron.harvester = harvester_class(database=self._patron.wagasci_database,
                                                     repository=self._history_location,
                                                     t2krun=None)

    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        if gr.id == "data":
            self._set_data_harvester(self._data_harvester_class)
            gr.xdata, gr.ydata = self._patron.gather_data()
        else:
            self._set_temp_harvester(self._temperature_harvester_class)
            gr.xdata, gr.ydata = self._patron.gather_data(detector_name=self._topology.get_enabled()[0].name)


class GainHistoryPlotterChannel(AdcCalibrationPlotterChannel):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(GainHistoryPlotterChannel, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.GainHarvesterChannel
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "Gain history of {};;Gain (ADC counts)".format(self._make_data_graph_title())

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_1d = "ALP"
        return super(GainHistoryPlotterChannel, self).setup_graphs()


class DarkNoiseHistoryPlotterChannel(AdcCalibrationPlotterChannel):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(DarkNoiseHistoryPlotterChannel, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.DarkNoiseHarvesterChannel
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "Dark noise rate history of {};;Dark noise rate (Hz)".format(self._make_data_graph_title())

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_1d = "ALP"
        return super(DarkNoiseHistoryPlotterChannel, self).setup_graphs()


class ThresholdHistoryPlotterChannel(AdcCalibrationPlotterChannel):

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        super(ThresholdHistoryPlotterChannel, self).__init__(*args, **kwargs)
        self._data_harvester_class = harvest.ThresholdHarvesterChannel
        self._temperature_harvester_class = harvest.AdcCalibrationTemperatureHarvester
        self._yrange = (0, 6)

    def set_title(self):
        # type: (...) -> None
        self._title = "Hit threshold level history of {};;Hit threshold (PEU)".format(self._make_data_graph_title())

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph]
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_1d = "ALP"
        return super(ThresholdHistoryPlotterChannel, self).setup_graphs()


class BcidPlotter(WagasciPlotter, ABC):

    def __init__(self, topology=None, *args, **kwargs):
        # type: (Optional[pl.topology.Topology], *Any, **Any) -> None
        super(BcidPlotter, self).__init__(*args, **kwargs)
        self._topology = topology if topology is not None else pl.topology.Topology(iterate_by_dif=True)
        self._wagasci_harvester_class = harvest.BcidHarvester

    def set_title(self):
        # type: (...) -> None
        self._title = "BCID history"
        super(BcidPlotter, self).set_title()
        self._title += ";;BCID counts (1 BCID count = 580 ns)"

    def setup_graphs(self):
        # type: (...) -> List[pl.graph.Graph2D]
        graphs = [graph.Graph2D("BCID")]
        graphs[0].raw_ydata = False
        self.plot_legend_flag = False
        self.xdata_is_datetime = True
        self.draw_options_2d = "COL"
        self.draw_options_2d += "Z"
        self.logscale = False
        return graphs

    def gather_data(self, gr):
        # type: (pl.graph.Graph) -> None
        if not self._topology.iterate_by_dif:
            raise RuntimeError("You must iterate through the active detectors by DIF")
        assert self._wagasci_harvester_class is not None, \
            "Derived class must set the _data_harvester_class attribute"
        enabled_detector = self._topology.get_enabled()[0]
        self._patron.harvester = self._wagasci_harvester_class(database=self._patron.wagasci_database,
                                                               repository=self._wagasci_repository,
                                                               t2krun=self._t2krun,
                                                               topology=self._topology)
        gr.xdata, gr.ydata = self._patron.gather_data(detector_name=enabled_detector.name, only_good=self._only_good)



import unittest

from wagascianpy.pyrmod.pyrmod import PyrameModule, PyrameSlowModule


class TestPyrameModule(unittest.TestCase):

    def setUp(self):
        self.module_name = "test"
        self.module_port = "TEST_PORT"
        self.function1 = "test1"
        self.arguments1 = ["arg1a", "arg1b"]
        self.result1 = (1, "result1")
        self.function0 = "test0"
        self.arguments0 = ["arg0a", "arg0b"]
        self.result0 = (0, "result0")
        self.simulation_dic = {self.function1: self.result1,
                               self.function0: self.result0}
        self.test = PyrameModule(module_name=self.module_name, simulate=True,
                                 simulation_dic=self.simulation_dic)

    def test_constructor(self):
        self.assertEqual(self.module_name, self.test.module_name)
        self.assertEqual(self.module_port, self.test.port_name)
        self.assertTrue(self.test.simulate)
        self.assertDictEqual(self.simulation_dic, self.test.simulation_dic)

    def test_execcmd1(self):
        self.assertEqual(self.result1, self.test.execcmd(self.function1, *self.arguments1))

    def test_execcmd0(self):
        self.assertRaises(RuntimeError, self.test.execcmd, self.function0, *self.arguments0)


class TestSlowModule(unittest.TestCase):

    def setUp(self):
        self.module_name = "test"
        self.module_port = "TEST_PORT"
        self.module_id = "gaaa"
        self.function1 = "test1"
        self.arguments1 = ["arg1a", "arg1b"]
        self.result1 = (1, "result1")
        self.function0 = "test0"
        self.arguments0 = ["arg0a", "arg0b"]
        self.result0 = (0, "result0")
        self.config_str = "a=aaa,b=bbb,c=ccc"
        self.simulation_dic = {self.function1: self.result1,
                               self.function0: self.result0}
        self.test = PyrameSlowModule(module_name=self.module_name,
                                     conf_string=self.config_str,
                                     simulate=True,
                                     simulation_dic=self.simulation_dic)

    def test_constructor(self):
        self.assertEqual(self.module_name, self.test.module_name)
        self.assertEqual(self.module_port, self.test.port_name)
        self.assertTrue(self.test.simulate)
        self.assertIn(self.module_name, self.test.module_id)
        self.assertDictEqual(self.simulation_dic, self.test.simulation_dic)

    def test_execcmd1(self):
        self.assertEqual(self.result1, self.test.execcmd(self.function1, *self.arguments1))

    def test_execcmd0(self):
        self.assertRaises(RuntimeError, self.test.execcmd, self.function0, *self.arguments0)

    def test_fixed_module_id(self):
        test_id = PyrameSlowModule(module_name=self.module_name,
                                   module_id=self.module_id,
                                   conf_string=self.config_str,
                                   simulate=True,
                                   simulation_dic=self.simulation_dic)
        self.assertEqual(self.module_id, test_id.module_id)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

"""

This module is used by the data harvesters to handle the decoded data TTrees for each DIF. Basically it just
opens a TTree and enables some of its TBranches. Optionally it can store a list of events for later retrieval.
It is not really meant to read the TTree itself, just open it and do some collateral checks.

The classes are organized in a tree-like structure where basically a "Detectors" object is just a list of "Detector"
objects, which in turn are a list of "Dif" objects. The detectors and DIFs are iterable like standard Python lists,
even if strictly speaking they are not subclasses of 'list'.

"""

from collections import namedtuple
from typing import Optional, List

try:
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    ROOT = None

from wagascianpy.plotting.topology import Topology, DetectorType, DifIndex, DetectorIndex, snakecase

WallmrdDifList = namedtuple('WallmrdDifList', ['top', 'bottom'])
WagasciDifList = namedtuple('WagasciDifList', ['top', 'side'])


class Dif(object):
    """
    Class representing a signle DIF. It contains the following info about a DIF:
      - DIF ID (a unique integer for each DIF)
      - A boolean flag if the DIF is enabled or disabled
      - A main TTree associated to the DIF (usually the raw data tree)
      - A list of TTrees that are friends with the main TTree
      - A list of spills (can be any Python list)
    """

    def __init__(self, name, dif_id, enabled):
        # type: (str, int, bool) -> None
        """
        :param name: name of the DIF
        :param dif_id: unique ID of the DIF
        :param enabled: true if the DIF is to be enabled
        """
        self.name = name
        self._dif_id = dif_id
        self._enabled = enabled
        self._tree_name = None
        self._chain = None
        self._friends = {}
        self._spills = None

    @property
    def dif_id(self):
        """
        :return: integer ID of the DIF
        """
        return self._dif_id

    def is_enabled(self):
        """
        :return: true if DIF is enabled
        """
        return self._enabled

    def enable(self):
        """
        Enable the DIF
        """
        self._enabled = True

    def disable(self):
        """
        Disable the DIF
        """
        self._enabled = False

    def add_tree(self, root_file, tree_name, tree_friends=None):
        # type: (str, str, Optional[List[str]]) -> None
        """
        Get the TTree handle from the ROOT file together with all its friends if any
        :param root_file: ROOT file where the TTree is stored
        :param tree_name: TTree name
        :param tree_friends: Names of the friend TTrees
        :return: nothing
        """
        if self._tree_name is None:
            self._tree_name = tree_name
        if tree_name != self._tree_name:
            raise ValueError("TTree name must be consistent among all ROOT files")
        if self._chain is None:
            self._chain = ROOT.TChain(tree_name)
        print("Tree {} found in {}".format(tree_name, root_file))
        self._chain.Add(root_file)
        if tree_friends:
            tfile = ROOT.TFile(root_file, "READ")
            for friend in tree_friends:
                if tfile.GetListOfKeys().Contains(friend):
                    print("Tree friend {} found in {}".format(friend, root_file))
                    if friend not in self._friends:
                        self._friends[friend] = ROOT.TChain(friend)
                        self._chain.AddFriend(self._friends[friend])
                    self._friends[friend].Add(root_file)
                else:
                    print("Tree friend {} not found in {}".format(friend, root_file))
            tfile.Close()

    def set_active_branches(self, active_branches):
        # type: (List[str]) -> None
        """
        Set which branches of the main TTree are active
        :param active_branches: names of the active branches
        :return: None
        """
        assert self._chain is not None, "Add a tree before accessing it"
        self._chain.SetBranchStatus("*", 0)
        for variable in active_branches:
            # if not any((branch.GetName() == variable for branch in self._chain.GetListOfBranches())):
            #     raise StopIteration("Branch {} not found in tree {}".format(variable, self._chain.GetName()))
            self._chain.SetBranchStatus(variable, 1)

    def get_tree(self):
        # type: (...) -> ROOT.TChain
        """
        :return: main TTree
        """
        assert self._chain is not None, "Add a tree before accessing it"
        return self._chain

    def get_friend(self, name):
        # type: (str) -> ROOT.TChain
        """
        :param name: name of friend TTree
        :return: a TTree whose friend with the main TTree
        """
        return self._friends[name]

    def has_friend(self, name):
        # type: (str) -> bool
        """
        :param name: name of friend TTree
        :return: true if the main TTree has a friend
        """
        return name in self._friends

    def set_spills(self, spills):
        # type: (List[sp.WagasciBsdSpill]) -> None
        """
        Store a list of spills
        :param spills: list of spills (but can be any list of objects)
        """
        assert isinstance(spills, list), "The spill list must be a list"
        self._spills = spills

    def get_spills(self):
        # type: (...) -> List[sp.WagasciBsdSpill]
        """
        :return: list of spills (but can be any list of objects)
        """
        assert self._spills is not None, "Please set the spill list before accessing it"
        return self._spills

    def has_spills(self):
        # type: (...) -> bool
        """
        :return: true if a non empty list of spills has been set
        """
        return bool(self._spills)

    def has_tree(self):
        # type: (...) -> bool
        """
        :return: true if the main TTree has been set
        """
        if self._chain is None:
            return False
        else:
            return True


class Detector(object):
    """
    Class representing a detector. It is useful to keep track of which detector is active and which not.
    If contains the following info:
      - Name of the detector
      - boolean Flag if the detector is enabled or not
      - List of Dif objects corresponding to the DIFs that make up the detector
    """

    def __init__(self, name, difs, enabled=True):
        # type: (str, Union[WallmrdDifList,WagasciDifList], bool) -> None
        """
        :param name: name of the detector
        :param difs: list of Dif objects
        :param enabled: true if the detector is to be enabled
        """
        self.name = name
        self._enabled = enabled
        if isinstance(difs, WallmrdDifList):
            self.detector_type = DetectorType.Wallmrd
            self.top = difs.top
            self.bottom = difs.bottom
            self.num_difs = 2
        elif isinstance(difs, WagasciDifList):
            self.detector_type = DetectorType.Wagasci
            self.top = difs.top
            self.side = difs.side
            self.num_difs = 2
        else:
            raise NotImplementedError("DIF list type {} not recognized".format(type(difs).__name__))

    def is_enabled(self):
        # type: (...) -> bool
        """
        :return: true if all Difs in the detector are enabled and have a TTree and spill list loaded, false otherwise
        """
        if not self._enabled:
            return False
        else:
            for dif in self:
                if not dif.has_tree() or not dif.has_spills():
                    return False
        return True

    def get_snake_case_name(self):
        """
        :return: name of the detector in snake case
        """
        return self.name.lower().replace(' ', '_')

    def __iter__(self):
        """ Returns the DifsIterator object """
        return DifsIterator(self)


class Detectors(object):
    """
    List of Detector objects representing all the detectors in the WAGASCI experiment.
    This class can be iterated by DIF or by detector. When iterated by DIF, the loop is over Dif objects,
    when iterated by detector, the loop is over the Detector objects.
    """

    def __init__(self, enabled_detectors=None):
        # type: (Optional[Topology]) -> None
        """
        :param enabled_detectors: which detectors are enabled and which are not
        """
        self._enabled_detectors = Topology(iterate_by_dif=True)
        if enabled_detectors is not None:
            self._enabled_detectors.import_topology(enabled_detectors)

        self.wallmrd_north = Detector(
            name=DetectorIndex.get_name(DetectorIndex.WallMrdNorth),
            difs=WallmrdDifList(Dif(name="top",
                                    dif_id=int(DifIndex.WallmrdNorthTop),
                                    enabled=self._enabled_detectors.wallmrd_north_top.is_enabled()),
                                Dif(name="bottom",
                                    dif_id=int(DifIndex.WallmrdNorthBottom),
                                    enabled=self._enabled_detectors.wallmrd_north_bottom.is_enabled())),
            enabled=(self._enabled_detectors.wallmrd_north_top.is_enabled() or
                     self._enabled_detectors.wallmrd_north_bottom.is_enabled()))
        self.wallmrd_south = Detector(
            name=DetectorIndex.get_name(DetectorIndex.WallMrdSouth),
            difs=WallmrdDifList(Dif(name="top",
                                    dif_id=int(DifIndex.WallmrdSouthTop),
                                    enabled=self._enabled_detectors.wallmrd_south_top.is_enabled()),
                                Dif(name="bottom",
                                    dif_id=int(DifIndex.WallmrdSouthBottom),
                                    enabled=self._enabled_detectors.wallmrd_south_bottom.is_enabled())),
            enabled=(self._enabled_detectors.wallmrd_south_top.is_enabled() or
                     self._enabled_detectors.wallmrd_south_bottom.is_enabled()))
        self.wagasci_upstream = Detector(
            name=DetectorIndex.get_name(DetectorIndex.WagasciUpstream),
            difs=WagasciDifList(Dif(name="top",
                                    dif_id=int(DifIndex.WagasciUpstreamTop),
                                    enabled=self._enabled_detectors.wagasci_upstream_top.is_enabled()),
                                Dif(name="side",
                                    dif_id=int(DifIndex.WagasciUpstreamSide),
                                    enabled=self._enabled_detectors.wagasci_upstream_side.is_enabled())),
            enabled=(self._enabled_detectors.wagasci_upstream_top.is_enabled() or
                     self._enabled_detectors.wagasci_upstream_side.is_enabled()))
        self.wagasci_downstream = Detector(
            name=DetectorIndex.get_name(DetectorIndex.WagasciDownstream),
            difs=WagasciDifList(Dif(name="top",
                                    dif_id=int(DifIndex.WagasciDownstreamTop),
                                    enabled=self._enabled_detectors.wagasci_downstream_top.is_enabled()),
                                Dif(name="side",
                                    dif_id=int(DifIndex.WagasciDownstreamSide),
                                    enabled=self._enabled_detectors.wagasci_downstream_side.is_enabled())),
            enabled=(self._enabled_detectors.wagasci_downstream_top.is_enabled() or
                     self._enabled_detectors.wagasci_downstream_side.is_enabled()))

    def __iter__(self):
        """ Returns the DetectorsIterator object """
        return DetectorsIterator(self)

    def get_dif(self, dif):
        # type: (Union[int, str]) -> Dif
        """
        :param dif: name or ID of the DIF
        :return: Dif object
        """
        dif_name = DifIndex.get_name_snakecase(dif) if isinstance(dif, int) else snakecase(dif)
        dif_id = DifIndex.get_index(dif) if isinstance(dif, str) else dif
        detector = next((detector for detector in self if snakecase(detector.name) in dif_name))
        return next(dif for dif in detector if dif.dif_id == dif_id)

    def get_detector(self, detector_id):
        # type: (Union[int, str]) -> Detector
        """
        :param detector_id: name or ID of the DETECTOR
        :return: Detector object whose ID is detector_id
        """
        if isinstance(detector_id, int):
            return getattr(self, DetectorIndex.get_name_snakecase(detector_id))
        else:
            return getattr(self, snakecase(detector_id))

    def has_dif(self, dif_id):
        # type: (...) -> bool
        """
        :param dif_id: unique DIF ID
        :return: true if the DIF is enabled and its TTree is loaded
        """
        try:
            dif = self.get_dif(dif_id)
        except (ValueError, AttributeError):
            return False
        else:
            return dif.is_enabled() and dif.has_tree()

    def disable_all_difs(self):
        # type: (...) -> None
        """
        Disable all DIFs
        """
        for detector in self:
            for dif in detector:
                dif.disable()

    def enable_dif(self, dif_id):
        # type: (int) -> None
        """
        :param dif_id: unique DIF ID
        :return: disable DIF whose ID is dif_id
        """
        dif_id = int(dif_id)
        self.get_dif(dif_id).enable()

    def activate_branches(self, active_branches_names):
        # type: (List[str]) -> None
        for detector in self:
            for dif in detector:
                if dif.has_tree():
                    dif.set_active_branches(active_branches=active_branches_names)
                else:
                    dif.disable()


class DifsIterator(object):
    """
    Itarate over all Dif objects
    """

    def __init__(self, detector):
        # Difs object reference
        self._detector = detector
        # member variable to keep track of current index
        self._index = 0

    def __next__(self):
        """ Returns the next value from difs object's lists """
        result = None
        if self._detector.detector_type == DetectorType.Wallmrd:
            if self._index < 2:
                if self._index == 0:
                    result = self._detector.top
                elif self._index == 1:
                    result = self._detector.bottom
                self._index += 1
                return result
        elif self._detector.detector_type == DetectorType.Wagasci:
            if self._index < 2:
                if self._index == 0:
                    result = self._detector.top
                elif self._index == 1:
                    result = self._detector.side
                self._index += 1
                return result
        else:
            raise NotImplementedError("Detector type {} not recognized".format(self._detector.detector_type))
        # End of Iteration
        raise StopIteration

    def next(self):
        return self.__next__()


class DetectorsIterator(object):
    """
    Iterate overall Detector objects
    """

    def __init__(self, detectors):
        # Detectors object reference
        self._detectors = detectors
        # member variable to keep track of current index
        self._index = 0

    def __next__(self):
        """ Returns the next value from detectors object's lists """
        result = None
        if self._index < len(DetectorIndex):
            if self._index == DetectorIndex.WallMrdNorth:
                result = self._detectors.wallmrd_north
            elif self._index == DetectorIndex.WallMrdSouth:
                result = self._detectors.wallmrd_south
            elif self._index == DetectorIndex.WagasciUpstream:
                result = self._detectors.wagasci_upstream
            elif self._index == DetectorIndex.WagasciDownstream:
                result = self._detectors.wagasci_downstream
            self._index += 1
            return result
        # End of Iteration
        raise StopIteration

    def next(self):
        return self.__next__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

# Python modules
import argparse
import textwrap

# user
import wagascianpy.analysis.mhistory2sqlite


###############################################################################
#                                  arguments                                  #
###############################################################################

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description=textwrap.dedent('''\
                                     Convert MIDAS history files .hst to SQLite format .sqlite3'''))

    PARSER.add_argument('-f', '--input-folder', metavar='<input folder>', dest='input_folder', type=str,
                        nargs=1, required=True, help=textwrap.dedent('''\
                        Path to a folder containing all the .hst files. If you want to convert a single file
                        please use the mh2sql utility bundled with MIDAS.'''))

    PARSER.add_argument('-r', '--recursive', dest='recursive', required=False, default=False, action='store_true',
                        help=textwrap.dedent('''\
                        If set the input directory is scanned recursively for .hst files.'''))

    PARSER.add_argument('-o', '--output-folder', metavar='<output folder>', dest='output_folder', type=str,
                        nargs=1, required=False, default=None, help=textwrap.dedent('''\
                        Output folder where the .sqlite3 files are saved (default is same as input)
                        '''))

    PARSER.add_argument('-a', '--start-time', metavar='<start time>', dest='start_time', type=str,
                        nargs=1, help='Start date and time in the format "%%Y/%%m/%%d %%H:%%M:%%S"', default=None,
                        required=False)

    PARSER.add_argument('-z', '--stop-time', metavar='<stop time>', dest='stop_time', type=str,
                        nargs=1, help='Stop date and time in the format "%%Y/%%m/%%d %%H:%%M:%%S"', default=None,
                        required=False)

    ARGS = PARSER.parse_args()

    if isinstance(ARGS.input_folder, list):
        ARGS.input_folder = ARGS.input_folder[0]
    INPUT_FOLDER = ARGS.input_folder
    if isinstance(ARGS.output_folder, list):
        ARGS.output_folder = ARGS.output_folder[0]
    OUTPUT_FOLDER = ARGS.output_folder
    if isinstance(ARGS.start_time, list):
        ARGS.start_time = ARGS.start_time[0]
    START_TIME = ARGS.start_time
    if isinstance(ARGS.stop_time, list):
        ARGS.stop_time = ARGS.stop_time[0]
    STOP_TIME = ARGS.stop_time
    RECURSIVE = ARGS.recursive

    wagascianpy.analysis.mhistory2sqlite.mhistory2sqlite(input_path=INPUT_FOLDER, output_folder=OUTPUT_FOLDER,
                                                         start_time=START_TIME, stop_time=STOP_TIME, recursive=RECURSIVE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ROOT

if __name__ == "__main__":
    # Path to the ROOT file
    # input_root_file = "/Users/akihiro/SynologyDrive/研究/WAGASCI/片山/analysis/data/physics_run_2020-02-11_22-23-18_106_ecal_dif_0_tree.root";
    input_root_file = "/Users/akihiro/SynologyDrive/研究/WAGASCI/片山/analysis/data/physics_run_2020-02-11_22-23-18_106_ecal_dif_4_tree.root";
    # TFile object
    tfile = ROOT.TFile(input_root_file, "READ")
    # get TTree from TFile
    ttree = tfile.Get("raw")
    # create canvas

    # Print image to file
    # create TCanvas of 1280x720 pixels
    tcanvas = ROOT.TCanvas("c1", "c1", 1280, 720)
    # ttree.Draw("charge[0][0][0]")
    # ttree.Draw("hit[0][0][0]")
    # tcanvas.Print("/Users/akihiro/Desktop/test2.png")

    # Loop example
    NCOLUMNS = 16
    NCHANNELS = 36
    NCHIPS = 20

    counter = 0
    
    histo = ROOT.TH2D("ch_vs_charge", "ch vs charge", 720, 0, 720, 500, 400, 900)
    for event in ttree:
        print(counter)
        for chip in range(int(NCHIPS)):
            for channel in range(int(NCHANNELS)):
                for column in range(int(NCOLUMNS)):
                    chipid = event.chipid[chip]
                    channelid = event.chanid[channel]
                    columnid = event.colid[column]
                    # columnid = 0
                    histo.Fill(channelid + NCHANNELS * chipid, event.charge[columnid + NCOLUMNS * channelid + NCOLUMNS * NCHANNELS * chipid])
#                    print(event.charge[column + NCOLUMNS * channel + NCOLUMNS * NCHANNELS * chip])

        # break after 10 spills to not have too long output
        counter += 1
        if counter == 100:
            break

    histo.Draw();
    tcanvas.Print("/Users/akihiro/Desktop/test2.png")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import os
import unittest

import wagascianpy.program.program
import wagascianpy.utils.utils


class TestProgram(unittest.TestCase):

    def setUp(self):
        self.bsd_database = os.path.abspath('./test_analyzer/bsddb.db')
        self.bsd_repository = '/media/neo/SDCARD/WAGASCI/BSD'
        self.topology_source = os.path.abspath(
            './test_analyzer/physics_run_2020-01-29_18-30-08_92/physics_run_2020-01-29_18-30-08_92.xml')
        self.sqlite_database = os.path.abspath('../databases/mh_temperature_sensors.sqlite3')
        self.history_location = os.path.abspath('./tmp_adc_calibration')

        self.run_dict = {}
        self.save_dict = {}
        self.run_location = './test_analyzer'
        self.save_location = './tmp_save'
        for run_name in os.listdir(self.run_location):
            if os.path.isdir(os.path.join(self.run_location, run_name)):
                self.run_dict[run_name] = os.path.abspath(os.path.join(self.run_location, run_name))
                self.save_dict[run_name] = os.path.abspath(os.path.join(self.save_location, run_name))

    def tearDown(self):
        for run_root_dir in self.run_dict.values():
            for filename in wagascianpy.utils.utils.find_files(run_root_dir, '.root') + \
                            wagascianpy.utils.utils.find_files(run_root_dir, '.txt') + \
                            wagascianpy.utils.utils.find_files(run_root_dir, '.png'):
                os.remove(filename)
        wagascianpy.utils.utils.silentremovedirs(self.save_location)
        wagascianpy.utils.utils.silentremovedirs(self.history_location)

    def test_program(self):
        program = wagascianpy.program.program.Program()
        program.set_run_location(self.run_dict)
        program.set_save_location(self.save_dict)
        program.multiple_runs_analyzer_save_location = self.history_location
        program.do_not_enforce_dependencies()
        program.stop_on_exception()
        program.add_step("decoder", overwrite_flag=True, compatibility_mode=False)
        program.add_step("spill_number_fixer")
        program.add_step("beam_summary_data",
                         bsd_database=self.bsd_database,
                         bsd_repository=self.bsd_repository,
                         t2krun=10,
                         recursive=True)
        program.add_step("temperature", sqlite_database=self.sqlite_database)
        program.add_step("adc_calibration", topology_source=self.topology_source)
        program.start()

    def test_program_builder_dependency(self):
        program = wagascianpy.program.program.Program()
        program.set_run_location(self.run_dict)
        program.set_save_location(self.save_dict)
        program.enforce_dependencies()
        self.assertRaises(RuntimeError, program.add_step, "spill_number_fixer")


if __name__ == '__main__':
    unittest.main()



import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock

import pyfakefs.fake_filesystem_unittest

# noinspection PyProtectedMember
import wagascianpy.viewer.utils


class TestCheckers(pyfakefs.fake_filesystem_unittest.TestCase):

    def setUp(self):
        self.setUpPyfakefs()
        self.fs.create_dir('/testdir')
        self.fs.create_file('/testdir/libwagasci.so')
        self.fs.create_dir('/testdir/run1')
        self.fs.create_dir('/testdir/run2')
        self.fs.create_dir('/testdir/run3')
        self.records = [{"name": "run1", "run_folder": "/testdir/run1"},
                        {"name": "run2", "run_folder": "/testdir/run2"},
                        {"name": "run3", "run_folder": "/testdir/run3"}]
        self.run_dic = {"run1": "/testdir/run1", "run2": "/testdir/run2", "run3": "/testdir/run3"}

    def test_check_repository_and_database_sanity_1(self):
        # should not throw
        repository = '/testdir'
        database = '/testdir/database.db'
        wagascianpy.viewer.utils.check_repository_and_database_sanity(repository=repository, database=database)
        repository = 'kekcc:/testdir'
        database = 'kekcc:/testdir/database.db'
        wagascianpy.viewer.utils.check_repository_and_database_sanity(repository=repository, database=database)
        repository = '/testdir'
        database = 'kekcc:/testdir/database.db'
        wagascianpy.viewer.utils.check_repository_and_database_sanity(repository=repository, database=database)
        repository = 'giorgio@kekcc:/testdir'
        database = 'giorgio@kekcc:/testdir/database.db'
        wagascianpy.viewer.utils.check_repository_and_database_sanity(repository=repository, database=database)

    def test_check_repository_and_database_sanity_2(self):
        repository = 'kekcc::/testdir'
        database = 'kekcc:/testdir/database.db'
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_repository_and_database_sanity,
                          repository=repository, database=database)

    def test_check_repository_and_database_sanity_3(self):
        repository = '/testdir'
        database = '/testdir/database.txt'
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_repository_and_database_sanity,
                          repository=repository, database=database)

    def test_check_repository_and_database_sanity_4(self):
        repository = ''
        database = '/testdir/database.db'
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_repository_and_database_sanity,
                          repository=repository, database=database)

    def test_check_repository_and_database_sanity_5(self):
        repository = '/testdir'
        database = ''
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_repository_and_database_sanity,
                          repository=repository, database=database)

    def test_check_repository_and_database_sanity_6(self):
        repository = ':/testdir'
        database = '/testdir/database.txt'
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_repository_and_database_sanity,
                          repository=repository, database=database)

    def test_check_wagasci_libdir_1(self):
        wagasci_libdir = '/testdir'
        self.assertEqual(wagasci_libdir, wagascianpy.viewer.utils.check_wagasci_libdir(wagasci_libdir))

    def test_check_wagasci_libdir_2(self):
        wagasci_libdir = ''
        self.assertIsNone(wagascianpy.viewer.utils.check_wagasci_libdir(wagasci_libdir))

    def test_check_wagasci_libdir_3(self):
        wagasci_libdir = None
        self.assertIsNone(wagascianpy.viewer.utils.check_wagasci_libdir(wagasci_libdir))

    def test_check_wagasci_libdir_4(self):
        wagasci_libdir = '/sdfaoigqpsaoJAADG'
        self.assertIsNone(wagascianpy.viewer.utils.check_wagasci_libdir(wagasci_libdir))

    def test_check_wagasci_libdir_5(self):
        wagasci_libdir = '/testdir/libwagasci.so'
        self.assertRaises(ValueError, wagascianpy.viewer.utils.check_wagasci_libdir, wagasci_libdir)

    def test_check_input_folder_1(self):
        self.assertRaises(OSError, wagascianpy.viewer.utils.check_input_folder,
                          input_folder='/asidsagjpoasdgopj', records=[], batch_mode=True)

    def test_check_input_folder_2(self):
        self.assertEqual(self.run_dic, wagascianpy.viewer.utils.check_input_folder(input_folder='/testdir',
                                                                                   records=self.records,
                                                                                   batch_mode=True))

    # noinspection PyUnusedLocal,PyShadowingBuiltins
    @mock.patch('wagascianpy.viewer.utils._get_input', return_value='/testdir')
    def test_check_input_folder_3(self, input):
        self.assertEqual(self.run_dic, wagascianpy.viewer.utils.check_input_folder(input_folder='',
                                                                                   records=self.records,
                                                                                   batch_mode=True))

    def test_check_input_folder_4(self):
        self.records.append({'name': 'run4', 'run_folder': '/testdir/run4'})
        self.assertEqual(self.run_dic, wagascianpy.viewer.utils.check_input_folder(input_folder='/testdir',
                                                                                   records=self.records,
                                                                                   batch_mode=True))

    def test_check_input_folder_5(self):
        del self.records[0]
        del self.run_dic["run1"]
        self.assertEqual(self.run_dic, wagascianpy.viewer.utils.check_input_folder(input_folder='/testdir',
                                                                                   records=self.records,
                                                                                   batch_mode=True))


if __name__ == '__main__':
    unittest.main()



import unittest

from unittest.mock import MagicMock, sentinel  # only Python3

from wagascianpy.utils.classproperty import classproperty
from wagascianpy.utils.classproperty import with_metaclass


class TestClassProperty(unittest.TestCase):
    def test_get_set(self):
        get_only_cls = MagicMock()
        get_set_get_cls = MagicMock()
        get_set_set_cls = MagicMock()

        # noinspection PyPropertyDefinition,PyMethodParameters
        @with_metaclass(classproperty.meta)
        class Z(object):
            _get_set = sentinel.nothing

            @classproperty
            def get_only(cls):
                get_only_cls(cls)
                return sentinel.get_only

            @classproperty
            def get_set(cls):
                get_set_get_cls(cls)
                return cls._get_set

            @get_set.setter
            def get_set(cls, value):
                get_set_set_cls(cls)
                cls._get_set = value

        for c, msg in [(Z, "class"), (Z(), "instance")]:
            with self.subTest(msg=msg):
                # Reset
                Z._get_set = sentinel.nothing

                # Test get_only
                self.assertEqual(sentinel.get_only, c.get_only)
                get_only_cls.assert_called_once_with(Z)
                get_only_cls.reset_mock()

                # Should return our initial "nothing" value
                self.assertEqual(sentinel.nothing, c.get_set)
                get_set_get_cls.assert_called_once_with(Z)
                get_set_get_cls.reset_mock()

                # Now test the set
                c.get_set = sentinel.get_set_val
                get_set_set_cls.assert_called_once_with(Z)
                get_set_set_cls.reset_mock()

                self.assertEqual(sentinel.get_set_val, c.get_set)
                get_set_get_cls.assert_called_once_with(Z)
                get_set_get_cls.reset_mock()

    def test_read_only(self):
        # noinspection PyMethodParameters
        @with_metaclass(classproperty.meta)
        class Z(object):
            _get_set = sentinel.nothing

            @classproperty
            def get_only(cls):
                return sentinel.get_only

        self.assertEqual(sentinel.get_only, Z.get_only)
        with self.assertRaises(AttributeError):
            Z.get_only = 123

    def test_proper_metaclass(self):
        # noinspection PyMethodParameters
        class Z(object):
            _get_set = sentinel.nothing

            @classproperty
            def get_only(cls):
                return sentinel.get_only

        with self.assertRaises(TypeError):
            self.assertEqual("should not resolve", Z.get_only)


if __name__ == "__main__":
    unittest.main()



import unittest

from mock import patch

from wagascianpy.utils.acq_config_xml import *
from wagascianpy.utils.topology import parse_topology_string, gdcc2dif_topology


class TestAcqConfigXml(unittest.TestCase):

    def setUp(self):
        self.acq_config_path = "/test.xml"
        self.topology_str = "full_setup"

    def test_wagasci_config(self):
        wg_config = WagasciConfig(simulate=True)
        self.assertEqual("wg_config", wg_config.module_name)
        self.assertEqual("WG_CONFIG_PORT", wg_config.port_name)
        self.assertIn("wg_config", wg_config.module_id)

    @patch('wagascianpy.utils.acq_config_xml.WagasciConfig')
    def test_create_acq_config_xml(self, mock_wagasci_config):
        conf_string = r'wg_config('
        conf_string += r'topology=%s,' % self.topology_str
        conf_string += r'acq_config_path=%s,' % self.acq_config_path
        conf_string += r'simulate=%s)' % (str(True))
        retcode, res = create_acq_config_xml(topology=self.topology_str, acq_config_path=self.acq_config_path,
                                             simulate=True)
        self.assertEqual(1, retcode)
        self.assertEqual("ok", res)
        mock_wagasci_config.assert_called_once_with(conf_string=conf_string)
        # might fail because of the conf_string ordering but it does not mean that the test failed

    @patch('wagascianpy.utils.acq_config_xml.WagasciConfig')
    def test_create_acq_config_xml_exception(self, mock_wagasci_config):
        mock_wagasci_config.side_effect = RuntimeError("test_exception")
        retcode, res = create_acq_config_xml(topology=self.topology_str, acq_config_path=self.acq_config_path,
                                             simulate=True)
        self.assertEqual(0, retcode)
        self.assertEqual("test_exception", res)


class TestTopology(unittest.TestCase):

    def setUp(self):
        self.wallmrd_south_before = "wallmrd_south"
        self.wallmrd_south_after = '{"1": {"3": {"1": 32, "3": 32, "2": 32}, "4": {"1": 32, "3": 32, "2": 32}}}'
        self.wagasci_downstream_before = "wagasci_downstream"
        self.wagasci_downstream_after = '{"2": {"3": {"11": 32, "10": 32, "13": 32, "12": 32, "15": 32, "14": 32, ' \
                                        '"17": 32, "16": 32, "19": 32, "18": 32, "20": 32, "1": 32, "3": 32, "2": 32, '\
                                        '"5": 32, "4": 32, "7": 32, "6": 32, "9": 32, "8": 32}, "4": {"11": 32, ' \
                                        '"10": 32, "13": 32, "12": 32, "15": 32, "14": 32, "17": 32, "16": 32, ' \
                                        '"19": 32, "18": 32, "20": 32, "1": 32, "3": 32, "2": 32, "5": 32, "4": 32, ' \
                                        '"7": 32, "6": 32, "9": 32, "8": 32}}}'
        self.custom_topology_before = r'{"1": {"3": {"1": 32\, "3": 32\, "2": 32}\, "4": {"1": 32\, "3": 32\, ' \
                                      r'"2": 32}}}'
        self.custom_topology_after = '{"1": {"3": {"1": 32, "3": 32, "2": 32}, "4": {"1": 32, "3": 32, "2": 32}}}'

    def test_topology_parser(self):
        self.assertEqual(self.wallmrd_south_after, parse_topology_string(self.wallmrd_south_before))
        self.assertEqual(self.wagasci_downstream_after, parse_topology_string(self.wagasci_downstream_before))
        self.assertEqual(self.custom_topology_after, parse_topology_string(self.custom_topology_before))

    def test_gdcc2dif_topology_dict(self):
        dif_topology = {"2": {"0": 32, "1": 32, "2": 32}, "3": {"0": 32, "1": 32, "2": 32}}
        self.assertEqual(dif_topology, gdcc2dif_topology(json.loads(self.wallmrd_south_after)))

    def test_gdcc2dif_topology_str(self):
        dif_topology = {"2": {"0": 32, "1": 32, "2": 32}, "3": {"0": 32, "1": 32, "2": 32}}
        self.assertEqual(dif_topology, gdcc2dif_topology(self.wallmrd_south_after))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

# Python modules
import argparse
import textwrap

# user
import wagascianpy.analysis.beam_summary_data

###############################################################################
#                                  arguments                                  #
###############################################################################

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description=textwrap.dedent('''\
                                     Add BSD info to the WAGASCI decoded file. The decoded file must contain
                                     the fixed_spill_number branch and all other branches created by the
                                     wgSpillNumberFixer program. '''))

    PARSER.add_argument('-f', '--input-file', metavar='<WAGASCI decoded file>', dest='input_file',
                        type=str, required=True, help=textwrap.dedent('''\
                        Path to a WAGASCI decoded file *_ecal_dif_X_tree.root or a directory containing those files.
                        If it is a directory all the files ending in *_ecal_dif_X_tree.root inside it will be analyzed.
                        If the "r" option is set the directory is traversed recursively.  
                        '''))

    PARSER.add_argument('-r', '--recursive', dest='recursive', required=False, default=False, action='store_true',
                        help=textwrap.dedent('''\
                        If set the input directory is scanned recursively for _ecal_dif_X_tree.root files.'''))

    PARSER.add_argument('-b', '--bsd-database', metavar='<BSD database>', dest='bsd_database', type=str,
                        required=True, help=textwrap.dedent('''\
                        Path to the Beam Summary Data database file. This file is created by the 
                        Wagasci Database Viewer (wagasciviewer.py) program and is usually called bsddb.db.
                        '''))
    PARSER.add_argument('-p', '--bsd-repository', metavar='<BSD repository>', dest='bsd_repository', type=str,
                        required=False, help=textwrap.dedent('''
                        Path to the Beam Summary Data local repository. This repository is just a folder 
                        containing all the BSD files. The repository must contain multiple subfolders each one 
                        named t2kXrun where X is the T2K run number.
                        '''))
    PARSER.add_argument('-t', '--t2krun', metavar='<T2K run>', dest='t2krun', type=int,
                        help='T2K run number (default is 10)', default=10, required=False)

    ARGS = PARSER.parse_args()

    INPUT_FILE = ARGS.input_file
    BSD_DATABASE = ARGS.bsd_database
    BSD_REPOSITORY = ARGS.bsd_repository
    T2KRUN = ARGS.t2krun
    RECURSIVE = ARGS.recursive

    wagascianpy.analysis.beam_summary_data.beam_summary_data(INPUT_FILE, BSD_DATABASE, BSD_REPOSITORY, T2KRUN,
                                                             RECURSIVE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio


import wagascianpy.program.program
import wagascianpy.program.program_builder
import wagascianpy.utils.environment


class Director:
    """
    The Director is only responsible for executing the building steps in a
    particular sequence. It is helpful when producing products according to a
    specific order or configuration. Strictly speaking, the Director class is
    optional, since the client can control builders directly.
    """

    def __init__(self):
        self._builder = None

    @property
    def builder(self):
        # type: (...) -> wagascianpy.program.program_builder.ProgramBuilder
        return self._builder

    @builder.setter
    def builder(self, builder):
        """
        The Director works with any builder instance that the client code passes
        to it. This way, the client code may alter the final type of the newly
        assembled product.
        """
        self._builder = builder

    """
    The Director can construct several product variations using the same
    building steps.
    """

    def build_decoder(self, overwrite_flag=False, compatibility_mode=False):
        self.builder.enforce_dependencies()
        self.builder.add_decoder(overwrite_flag=overwrite_flag, compatibility_mode=compatibility_mode)

    def build_make_hist(self, overwrite):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_make_hist(overwrite=overwrite)

    def build_bcid_distribution(self, chip_by_chip, overwrite):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_bcid_distribution(chip_by_chip=chip_by_chip, overwrite=overwrite)

    def build_adc_distribution(self, chip_by_chip, overwrite):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_bcid_distribution(chip_by_chip=chip_by_chip, overwrite=overwrite)

    def build_tdc_calibration(self):
        self.builder.do_not_enforce_dependencies()
        self.builder.tdc_calibration()

    def build_spill_number_fixer(self):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_spill_number_fixer(enable_graphics=False)

    def build_apply_detector_flags(self, wagasci_database_location):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_apply_detector_flags(wagasci_database=wagasci_database_location)

    def build_sanity_check(self, wagasci_database_location):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_sanity_check(wagasci_database=wagasci_database_location)

    def build_beam_summary_data(self,
                                bsd_database_location,
                                bsd_repository_location,
                                download_bsd_database_location="/tmp/bsd/bsddb.db",
                                download_bsd_repository_location="/tmp/bsd",
                                t2krun=10):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_beam_summary_data(bsd_database_location=bsd_database_location,
                                           bsd_repository_location=bsd_repository_location,
                                           download_bsd_database_location=download_bsd_database_location,
                                           download_bsd_repository_location=download_bsd_repository_location,
                                           t2krun=t2krun,
                                           recursive=True)

    def build_temperature(self, sqlite_database):
        self.builder.do_not_enforce_dependencies()
        self.builder.add_temperature(sqlite_database=sqlite_database)

    def build_adc_calibration(self, history_location=None):
        self.builder.do_not_enforce_dependencies()
        if not history_location:
            try:
                env = wagascianpy.utils.environment.WagasciEnvironment()
                history_location = env["WAGASCI_PEUDIR"]
            except KeyError as error:
                raise KeyError("Could not find WAGASCI history directory : {}".format(str(error)))
        self.builder.multiple_runs_analyzer_save_location = history_location
        self.builder.add_adc_calibration()

    def build_all(self,
                  wagasci_database_location,
                  bsd_database_location,
                  bsd_repository_location,
                  history_location,
                  sqlite_database,
                  download_bsd_database_location="/tmp/bsd/bsddb.db",
                  download_bsd_repository_location="/tmp/bsd",
                  t2krun=10,
                  overwrite_flag=False,
                  compatibility_mode=False
                  ):
        self.build_decoder(overwrite_flag=overwrite_flag, compatibility_mode=compatibility_mode)
        self.build_tdc_calibration()
        self.build_spill_number_fixer()
        self.build_beam_summary_data(bsd_database_location=bsd_database_location,
                                     bsd_repository_location=bsd_repository_location,
                                     download_bsd_database_location=download_bsd_database_location,
                                     download_bsd_repository_location=download_bsd_repository_location,
                                     t2krun=t2krun)
        self.build_temperature(sqlite_database=sqlite_database)
        self.build_adc_calibration(history_location=history_location)
        self.build_apply_detector_flags(wagasci_database_location=wagasci_database_location)
        self.build_sanity_check(wagasci_database_location=wagasci_database_location)
        self.builder.enforce_dependencies()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
import os
import pprint

try:
    # Python 2 only:
    import ConfigParser as cp
except ImportError:
    # Python 2 and 3 (after ``pip install configparser``)
    import configparser as cp
import dependency_injector.containers
import dependency_injector.providers
import inflection
from enum import Enum
from six import string_types
from typing import Optional, Dict, List, Any, Tuple, Iterable

import wagascianpy.utils.environment
import wagascianpy.utils.utils
from wagascianpy.utils.classproperty import classproperty

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3


class RepositoryType(Enum):
    Simple = 1
    Borg = 2


###############################################################################################
#                                        Default values                                       #
###############################################################################################

_DEFAULT_CONFIGURATION_FILENAME = "anpy_conf.ini"

# WAGASCI database configuration
_WAGASCI_DATABASE = "kekcc:/hsm/nu/wagasci/data/run10/database/wagascidb.db"
_WAGASCI_REPOSITORY = "kekcc:/hsm/nu/wagasci/data/run10/wagasci/rawdata"
_WAGASCI_REPOSITORY_TYPE = RepositoryType.Simple
_WAGASCI_DOWNLOAD_LOCATION = "/tmp/download"
_WAGASCI_DECODED_LOCATION = "/tmp/decoded"
_WAGASCI_DEFAULT_DATABASE_NAME = "wagascidb.db"

# BSD database configuration
_BSD_DATABASE = "kekcc:/hsm/nu/wagasci/data/run10/database/bsddb.db"
_BSD_REPOSITORY = "kekcc:/gpfs/fs03/t2k/beam/exp/data/beam_summary/current/t2krun10"
_BSD_DOWNLOAD_LOCATION = "/tmp/bsd"
_BSD_DEFAULT_DATABASE_NAME = "bsddb.db"

# Global configuration
_T2KRUN = 10
_HISTORY_LOCATION = "/tmp/history"
_WAGASCI_LIBDIR = "/usr/local/lib/wagasci/calibration"

# Temperature database configuration
TEMPERATURE_SQLITE_DATABASE = "kekcc:/hsm/nu/wagasci/data/run10/database/mh_temperature_sensors_t2krun10.sqlite3"


###############################################################################################
#                                         Utilities                                           #
###############################################################################################

def conf_file_finder(filename, project_path=None):
    # type: (str, Optional[str, List[str]]) -> str
    """
    Four types of configuration files are considered in the following priority (higher is preferred). If the
    configuration file is not found in the top priority folder the next one in priority is searched. If no file is
    found the 2) User configuration file is returned even if it does not exists.
     1 - Argument : The file path is passed as argument
     2 - Environment : The environment variable WAGASCI_CONFDIR is assumed to be the path to folder containing the
                     configuration file.
     3 - User        : The configuration folder is $HOME/.wagasci
     4 - System      : The configuration folder is /usr/local/wagasci (Only UNIX)
     5 - Project     : Configuration file used during development and stored in the project folder/folders.
                     "filename" is appended to each "project_path" folder.
    :param filename: name of the file
    :param project_path: project paths containing the configuration files (only used during development)
    :return: full path to the configuration file
    """
    home = os.path.expanduser("~")
    wagasci_environ = None

    # Read the environment variable
    try:
        env = wagascianpy.utils.environment.WagasciEnvironment()
        wagasci_environ = os.path.join(env["WAGASCI_CONFDIR"], filename)
    except KeyError:
        pass

    # Append the filename to the project paths
    if isinstance(project_path, list):
        wagasci_project = list(map(lambda x: os.path.join(x, filename), project_path))
    elif isinstance(project_path, str):
        wagasci_project = [os.path.join(project_path, filename)]
    else:
        wagasci_project = None

    # User configuration file
    wagasci_user = os.path.join(home, '.wagasci', filename)

    # System configuration file
    wagasci_system = os.path.join('/usr/local/wagasci/', filename)

    conf_file = None
    # 1 : Argument
    if os.path.exists(filename):
        conf_file = filename
    # 2 : Environment
    elif wagasci_environ and os.path.exists(wagasci_environ):
        conf_file = wagasci_environ
    # 3 : User
    elif wagasci_user and os.path.exists(wagasci_user):
        conf_file = wagasci_user
    # 4 : System
    elif wagasci_system and os.path.exists(wagasci_system):
        conf_file = wagasci_system
    # 5 : Project
    elif wagasci_project:
        for path in wagasci_project:
            if os.path.exists(path):
                conf_file = path
                break
    # No file is found (assuming user configuration file)
    else:
        wagascianpy.utils.utils.mkdir_p(os.path.join(home, '.wagasci'))
        conf_file = wagasci_user
    return conf_file


def _conf_setter_raise(cls, name, value):
    # type: (Any, str, Any) -> None
    raise ValueError("Invalid value {} for option {} of class {}".format(name, value, cls.__name__))


def conf_setter(cls, name, value):
    # type: (Any, str, Any) -> None
    try:
        if isinstance(getattr(cls, name), Enum):
            if isinstance(value, Enum):
                setattr(cls, name, value)
            elif isinstance(value, IntTypes):
                setattr(cls, name, type(getattr(cls, name))(value))
            elif isinstance(value, string_types):
                if '.' in value:
                    value = value.split('.')[1]
                try:
                    setattr(cls, name, getattr(type(getattr(cls, name)), inflection.camelize(value.lower())))
                except AttributeError:
                    raise ValueError
        elif isinstance(getattr(cls, name), bool):
            if isinstance(value, string_types):
                value = value.lower()
            if value in [0, False, None, 'none', 'false', 'off', 'no']:
                setattr(cls, name, False)
            elif value in [1, True, 'true', 'on', 'yes']:
                setattr(cls, name, True)
            else:
                raise ValueError
        elif isinstance(getattr(cls, name), int):
            value = int(value) if value is not None else 0
            setattr(cls, name, int(value))
        elif isinstance(getattr(cls, name), float):
            value = float(value) if value is not None else 0.
            setattr(cls, name, float(value))
        else:
            value = str(value) if value is not None else ''
            setattr(cls, name, str(value))
    except ValueError:
        _conf_setter_raise(cls, name, value)


def conf_getter(cls, name):
    # type: (Any, str) -> Any
    return getattr(cls, name)


###############################################################################################
#                                    Configuration classes                                    #
###############################################################################################

@wagascianpy.utils.classproperty.with_metaclass(classproperty.meta)
class VirtualConfiguration(object):

    @classmethod
    def to_dict(cls):
        # type: (...) -> Dict[str, Any]
        return {key: val.__get__(instance=None, owner=cls) for key, val in vars(cls).items()
                if isinstance(val, classproperty)}

    @classmethod
    def get_hidden_vars(cls):
        # type: (...) -> Iterable[str]
        return [key for key, val in vars(cls).items()
                if key.startswith('_') and not key.startswith('__') and not callable(val)]

    @classmethod
    def get_classproperties(cls):
        # type: (...) -> Iterable[str]
        return [key for key, val in vars(cls).items() if isinstance(val, classproperty)]

    @classmethod
    def dump(cls, indentation_level=0):
        # type: (int) -> None
        for propery in cls.get_classproperties():
            print("{}{} = {}".format(' ' * indentation_level, propery, getattr(cls, propery)))


# noinspection PyMethodParameters,PyPropertyDefinition
class GlobalConfiguration(VirtualConfiguration):
    _t2krun = _T2KRUN
    _history_location = _HISTORY_LOCATION
    _wagasci_libdir = _WAGASCI_LIBDIR

    @classmethod
    def reset(cls):
        cls._t2krun = _T2KRUN
        cls._history_location = _HISTORY_LOCATION

    @classmethod
    def is_volatile(cls):
        return False

    @classproperty
    def t2krun(cls):
        return conf_getter(cls=cls, name='_t2krun')

    @t2krun.setter
    def t2krun(cls, value):
        conf_setter(cls=cls, name='_t2krun', value=value)

    @classproperty
    def history_location(cls):
        return conf_getter(cls=cls, name='_history_location')

    @history_location.setter
    def history_location(cls, value):
        conf_setter(cls=cls, name='_history_location', value=value)

    @classproperty
    def wagasci_libdir(cls):
        return conf_getter(cls=cls, name='_wagasci_libdir')

    @wagasci_libdir.setter
    def wagasci_libdir(cls, value):
        conf_setter(cls=cls, name='_wagasci_libdir', value=value)


# noinspection PyMethodParameters,PyPropertyDefinition
class Wagascidb(VirtualConfiguration):
    _wagasci_database = _WAGASCI_DATABASE
    _wagasci_repository = _WAGASCI_REPOSITORY
    _repository_type = _WAGASCI_REPOSITORY_TYPE
    _wagasci_download_location = _WAGASCI_DOWNLOAD_LOCATION
    _wagasci_decoded_location = _WAGASCI_DECODED_LOCATION
    _default_database_name = _WAGASCI_DEFAULT_DATABASE_NAME

    @classmethod
    def reset(cls):
        cls._wagasci_database = _WAGASCI_DATABASE
        cls._wagasci_repository = _WAGASCI_REPOSITORY
        cls._repository_type = _WAGASCI_REPOSITORY_TYPE
        cls._wagasci_download_location = _WAGASCI_DOWNLOAD_LOCATION
        cls._wagasci_decoded_location = _WAGASCI_DECODED_LOCATION
        cls._default_database_name = _WAGASCI_DEFAULT_DATABASE_NAME

    @classmethod
    def is_volatile(cls):
        return False

    @classproperty
    def wagasci_database(cls):
        return conf_getter(cls=cls, name='_wagasci_database')

    @wagasci_database.setter
    def wagasci_database(cls, value):
        conf_setter(cls=cls, name='_wagasci_database', value=value)

    @classproperty
    def wagasci_repository(cls):
        return conf_getter(cls=cls, name='_wagasci_repository')

    @wagasci_repository.setter
    def wagasci_repository(cls, value):
        conf_setter(cls=cls, name='_wagasci_repository', value=value)

    @classproperty
    def repository_type(cls):
        return conf_getter(cls=cls, name='_repository_type')

    @repository_type.setter
    def repository_type(cls, value):
        conf_setter(cls=cls, name='_repository_type', value=value)

    @classproperty
    def wagasci_download_location(cls):
        return conf_getter(cls=cls, name='_wagasci_download_location')

    @wagasci_download_location.setter
    def wagasci_download_location(cls, value):
        conf_setter(cls=cls, name='_wagasci_download_location', value=value)

    @classproperty
    def wagasci_decoded_location(cls):
        return conf_getter(cls=cls, name='_wagasci_decoded_location')

    @wagasci_decoded_location.setter
    def wagasci_decoded_location(cls, value):
        conf_setter(cls=cls, name='_wagasci_decoded_location', value=value)

    @classproperty
    def default_database_name(cls):
        return conf_getter(cls=cls, name='_default_database_name')

    @default_database_name.setter
    def default_database_name(cls, value):
        conf_setter(cls=cls, name='_default_database_name', value=value)


# noinspection PyMethodParameters,PyPropertyDefinition
class Bsddb(VirtualConfiguration):
    _bsd_database = _BSD_DATABASE
    _bsd_repository = _BSD_REPOSITORY
    _bsd_download_location = _BSD_DOWNLOAD_LOCATION
    _default_database_name = _BSD_DEFAULT_DATABASE_NAME

    @classmethod
    def reset(cls):
        cls._bsd_database = _BSD_DATABASE
        cls._bsd_repository = _BSD_REPOSITORY
        cls._bsd_download_location = _BSD_DOWNLOAD_LOCATION
        cls._default_database_name = _BSD_DEFAULT_DATABASE_NAME

    @classmethod
    def is_volatile(cls):
        return False

    @classproperty
    def bsd_database(cls):
        return conf_getter(cls=cls, name='_bsd_database')

    @bsd_database.setter
    def bsd_database(cls, value):
        conf_setter(cls=cls, name='_bsd_database', value=value)

    @classproperty
    def bsd_repository(cls):
        return conf_getter(cls=cls, name='_bsd_repository')

    @bsd_repository.setter
    def bsd_repository(cls, value):
        conf_setter(cls=cls, name='_bsd_repository', value=value)

    @classproperty
    def bsd_download_location(cls):
        return conf_getter(cls=cls, name='_bsd_download_location')

    @bsd_download_location.setter
    def bsd_download_location(cls, value):
        conf_setter(cls=cls, name='_bsd_download_location', value=value)

    @classproperty
    def default_database_name(cls):
        return conf_getter(cls=cls, name='_default_database_name')

    @default_database_name.setter
    def default_database_name(cls, value):
        conf_setter(cls=cls, name='_default_database_name', value=value)


# noinspection PyMethodParameters,PyMethodParameters,PyPropertyDefinition
class Temperature(VirtualConfiguration):
    _temperature_sqlite_database = TEMPERATURE_SQLITE_DATABASE

    @classmethod
    def reset(cls):
        cls._temperature_sqlite_database = TEMPERATURE_SQLITE_DATABASE

    @classmethod
    def is_volatile(cls):
        return False

    @classproperty
    def temperature_sqlite_database(cls):
        return conf_getter(cls=cls, name='_temperature_sqlite_database')

    @temperature_sqlite_database.setter
    def temperature_sqlite_database(cls, value):
        conf_setter(cls=cls, name='_temperature_sqlite_database', value=value)


###############################################################################################
#                                    Configuration parser                                     #
###############################################################################################


class WagasciConfigParser(GlobalConfiguration,
                          Wagascidb,
                          Bsddb,
                          Temperature):

    def _baseclasses(self):
        # type: (...) -> Optional[Tuple]
        base_classes = WagasciConfigParser.__bases__
        if self.__class__ != WagasciConfigParser:
            base_classes += tuple([bc for bc in self.__class__.__bases__ if bc != WagasciConfigParser])
        return base_classes

    def __init__(self, config_file_path=None):
        # type: (Optional[str]) -> None

        # Create ConfigParser default object
        self._config_parser = cp.ConfigParser()

        # Set default values and create the section object attributes
        for base_class in self._baseclasses():
            section_name = inflection.underscore(base_class.__name__)
            setattr(self, section_name, base_class)

        # fill the ConfigParser object
        self._fill_config_parser()

        # check file access permissions
        if config_file_path:
            if not os.access(os.path.dirname(config_file_path), os.W_OK):
                raise OSError("The configuration file path is not writable "
                              "by the current user : %s" % config_file_path)
            # If the configuration file exists read it
            if os.path.exists(config_file_path):
                print('Reading configuration from file "{}"'.format(config_file_path))
                self._config_parser.read(config_file_path)
                for section in self._config_parser.sections():
                    if not hasattr(self, section):
                        raise AttributeError("Section name not recognized : %s" % section)
                    for key, val in self._config_parser.items(section=section):
                        setattr(getattr(self, section), key, val)
            else:
                # If the file does not exist write the default configuration to file
                with open(config_file_path, 'w') as configfile:
                    print('Writing default configuration to file "{}"'.format(config_file_path))
                    self._config_parser.write(configfile)

    def get_sections(self):
        # type: (...) -> Optional[List[str]]
        return [key for key in vars(self) if not key.startswith('_')]

    def get_section(self, name):
        # type: (str) -> Any
        return getattr(self, name)

    def prettyprint(self):
        for section in self.get_sections():
            print()
            print(section.upper().replace('_', ' '))
            pprint.pprint(self.get_section(section).to_dict())

    def _fill_config_parser(self):
        for section in self.get_sections():
            if not self.get_section(section).is_volatile():
                if not self._config_parser.has_section(section):
                    self._config_parser.add_section(section)
                for key, var in self.get_section(section).to_dict().items():
                    if isinstance(var, Enum):
                        var = str(var).split('.')[1]
                    self._config_parser.set(section, key, str(var))

    def write(self, config_file_path):
        self._fill_config_parser()
        with open(config_file_path, 'w') as configfile:
            self._config_parser.write(configfile)


###############################################################################################
#                                Configuration Global Object                                  #
###############################################################################################


def _create_sections(cls, parser_class=WagasciConfigParser):
    conf_file = conf_file_finder(filename=_DEFAULT_CONFIGURATION_FILENAME)
    print("Using INI configuration file %s" % conf_file)
    parser = parser_class(conf_file)
    setattr(cls, 'parser', parser)
    for section in parser.get_sections():
        if not hasattr(cls, section):
            setattr(cls, section, dependency_injector.providers.Configuration(section))
            getattr(cls, section).override(getattr(parser, section).to_dict())
    return cls


# wrap _create_sections to allow for deferred calling
def create_sections(cls=None, parser_class=WagasciConfigParser):
    if cls:
        return _create_sections(cls)
    else:
        def wrapper(function):
            return _create_sections(function, parser_class)

        return wrapper


@create_sections
class Configuration(dependency_injector.containers.DeclarativeContainer):
    """IoC container of configuration providers."""

    @classmethod
    def create_section(cls, name):
        # type: (str) -> None
        if not hasattr(cls, name):
            setattr(cls, name, dependency_injector.providers.Configuration(name))

    @classmethod
    def delete_section(cls, name):
        # type: (str) -> None
        if hasattr(cls, name):
            delattr(cls, name)

    @classmethod
    def get_sections(cls):
        # type: (...) -> Optional[List[str]]
        return [val.get_name() for val in vars(cls).values()
                if isinstance(val, dependency_injector.providers.Configuration)]

    @classmethod
    def dump(cls):
        for section in cls.parser.get_sections():
            print("[{}]".format(section))
            getattr(cls.parser, section).dump(indentation_level=2)


if __name__ == "__main__":
    Configuration.parser.prettyprint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

from enum import Enum

from six import string_types
import wagascianpy.plotting.detector

import ROOT
ROOT.PyConfig.IgnoreCommandLineOptions = True


class Colors(Enum):
    White = ROOT.kWhite
    Red = ROOT.kRed
    AlternateRed = ROOT.kRed + 2
    Green = ROOT.kGreen
    AlternateGreen = ROOT.kGreen + 2
    Blue = ROOT.kBlue
    AlternateBlue = ROOT.kBlue - 2
    Orange = ROOT.kOrange
    AlternateOrange = ROOT.kOrange + 2
    Cyan = ROOT.kCyan
    AlternateCyan = ROOT.kCyan + 2
    Azure = ROOT.kAzure
    AlternateAzuer = ROOT.kAzure + 2
    Black = ROOT.kBlack

    @classmethod
    def get_by_dif(cls, dif_id):
        dif_id = int(dif_id)
        if dif_id == 0:
            return cls.get_by_detector("WallMRD north top")
        elif dif_id == 1:
            return cls.get_by_detector("WallMRD north bottom")
        elif dif_id == 2:
            return cls.get_by_detector("WallMRD south top")
        elif dif_id == 3:
            return cls.get_by_detector("WallMRD south bottom")
        elif dif_id == 4:
            return cls.get_by_detector("WAGASCI upstream top")
        elif dif_id == 5:
            return cls.get_by_detector("WAGASCI upstream side")
        elif dif_id == 6:
            return cls.get_by_detector("WAGASCI downstream top")
        elif dif_id == 7:
            return cls.get_by_detector("WAGASCI downstream side")
        return cls.Black

    @classmethod
    def get_by_detector(cls, detector):
        name = ""
        if isinstance(detector, wagascianpy.plotting.detector.Detector):
            name = detector.name
        elif isinstance(detector, string_types):
            name = detector
        else:
            TypeError("Wrong type : {}".format(type(detector).__name__))
        if name == "WallMRD north" or name == "WallMRD north top":
            return cls.Green
        elif name == "WallMRD north bottom":
            return cls.AlternateGreen
        elif name == "WallMRD south" or name == "WallMRD south top":
            return cls.Cyan
        elif name == "WallMRD south bottom":
            return cls.AlternateCyan
        elif name == "WAGASCI upstream" or name == "WAGASCI upstream top":
            return cls.Orange
        elif name == "WAGASCI upstream side":
            return cls.AlternateOrange
        elif name == "WAGASCI downstream" or name == "WAGASCI downstream top":
            return cls.Blue
        elif name == "WAGASCI downstream side":
            return cls.AlternateBlue
        return cls.Black



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import os
import unittest

import wagascianpy.analysis.analyzer
import wagascianpy.utils.acq_config_xml
import wagascianpy.utils.utils


class TestAnalyzer(unittest.TestCase):

    def setUp(self):
        self.run_root_dir = os.path.abspath('./test_analyzer/physics_run_2020-01-29_18-30-08_92')
        self.bsd_database = os.path.abspath('./test_analyzer/bsddb.db')
        self.bsd_repository = "/media/neo/SDCARD/WAGASCI/BSD"
        self.run_name = 'physics_run_2020-01-29_18-30-08_92'
        self.run_number = 92
        self.save_location = os.path.abspath('./tmp_save')
        self.history_location = os.path.abspath('./tmp_adc_calibration')
        self.sqlite_database = os.path.abspath('../databases/mh_temperature_sensors.sqlite3')
        self.wagasci_xml_config = wagascianpy.utils.acq_config_xml.acqconfigxml_file_finder(
            self.run_root_dir, self.run_name)
        self.chains = {self.run_name: {}}

    def tearDown(self):
        for filename in wagascianpy.utils.utils.find_files(self.run_root_dir, '.root') + \
                        wagascianpy.utils.utils.find_files(self.run_root_dir, '.txt') + \
                        wagascianpy.utils.utils.find_files(self.run_root_dir, '.png'):
            os.remove(filename)
        wagascianpy.utils.utils.silentremovedirs(self.save_location)
        wagascianpy.utils.utils.silentremovedirs(self.history_location)

    def test_decoder(self):
        decoder = wagascianpy.analysis.analyzer.Decoder(run_name=self.run_name,
                                                        run_number=self.run_number,
                                                        run_root_dir=self.run_root_dir,
                                                        output_dir=self.save_location,
                                                        overwrite_flag=True,
                                                        compatibility_mode=False)

        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)

    def test_adc_distribution(self):
        decoder = wagascianpy.analysis.analyzer.Decoder(run_name=self.run_name,
                                                        run_number=self.run_number,
                                                        run_root_dir=self.run_root_dir,
                                                        output_dir=self.save_location,
                                                        overwrite_flag=True,
                                                        compatibility_mode=False)
        adc = wagascianpy.analysis.analyzer.AdcDistribution(run_name=self.run_name,
                                                            run_number=self.run_number,
                                                            run_root_dir=self.save_location,
                                                            output_dir=self.save_location,
                                                            chip_by_chip=True)
        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains[self.run_name] = {}
        adc.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)

    def test_spill_number_fixer(self):
        decoder = wagascianpy.analysis.analyzer.Decoder(run_name=self.run_name,
                                                        run_number=self.run_number,
                                                        run_root_dir=self.run_root_dir,
                                                        output_dir=self.save_location,
                                                        overwrite_flag=True,
                                                        compatibility_mode=False)
        spill_fixer = wagascianpy.analysis.analyzer.SpillNumberFixer(run_name=self.run_name,
                                                                     run_root_dir=self.save_location,
                                                                     output_dir=self.save_location)

        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains[self.run_name] = {}
        spill_fixer.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)

    def test_beam_summary_data(self):
        decoder = wagascianpy.analysis.analyzer.Decoder(run_name=self.run_name,
                                                        run_number=self.run_number,
                                                        run_root_dir=self.run_root_dir,
                                                        output_dir=self.save_location,
                                                        overwrite_flag=True,
                                                        compatibility_mode=False)
        spill_fixer = wagascianpy.analysis.analyzer.SpillNumberFixer(run_name=self.run_name,
                                                                     run_root_dir=self.save_location,
                                                                     output_dir=self.save_location)
        bsd = wagascianpy.analysis.analyzer.BeamSummaryData(run_name=self.run_name,
                                                            run_number=self.run_number,
                                                            run_root_dir=self.save_location,
                                                            output_dir=self.save_location,
                                                            bsd_database=self.bsd_database,
                                                            bsd_repository=self.bsd_repository,
                                                            t2krun=10,
                                                            recursive=True)
        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains[self.run_name] = {}
        spill_fixer.spawn(self.chains[self.run_name])
        bsd.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)

    def test_adc_calibration(self):
        decoder = wagascianpy.analysis.analyzer.Decoder(run_name=self.run_name,
                                                        run_number=self.run_number,
                                                        run_root_dir=self.run_root_dir,
                                                        output_dir=self.save_location,
                                                        overwrite_flag=True,
                                                        compatibility_mode=False)
        spill_fixer = wagascianpy.analysis.analyzer.SpillNumberFixer(run_name=self.run_name,
                                                                     run_root_dir=self.save_location,
                                                                     output_dir=self.save_location)
        bsd = wagascianpy.analysis.analyzer.BeamSummaryData(run_name=self.run_name,
                                                            run_number=self.run_number,
                                                            run_root_dir=self.save_location,
                                                            output_dir=self.save_location,
                                                            bsd_database=self.bsd_database,
                                                            bsd_repository=self.bsd_repository,
                                                            t2krun=10,
                                                            recursive=True)
        temp = wagascianpy.analysis.analyzer.Temperature(run_name=self.run_name,
                                                         run_number=self.run_number,
                                                         run_root_dir=self.save_location,
                                                         output_dir=self.save_location,
                                                         sqlite_database=self.sqlite_database)
        adc_calibration = wagascianpy.analysis.analyzer.AdcCalibration(run_name="week",
                                                                       run_number=42,
                                                                       run_root_dir=self.save_location,
                                                                       output_dir=self.history_location,
                                                                       topology_source=self.wagasci_xml_config)

        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains[self.run_name] = {}
        spill_fixer.spawn(self.chains[self.run_name])
        bsd.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains[self.run_name] = {}
        temp.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)
        self.chains["week"] = {}
        adc_calibration.spawn(self.chains["week"])
        wagascianpy.utils.utils.join_chains(self.chains)

    def test_analyzer_factory(self):
        producer = wagascianpy.analysis.analyzer.AnalyzerFactoryProducer()
        factory = producer.get_factory("decoder",
                                       overwrite_flag=True,
                                       compatibility_mode=False)

        decoder = factory.get_analyzer(run_name=self.run_name,
                                       run_number=self.run_number,
                                       run_root_dir=self.run_root_dir,
                                       output_dir=self.save_location)

        decoder.spawn(self.chains[self.run_name])
        wagascianpy.utils.utils.join_chains(self.chains)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ROOT

if __name__ == "__main__":
    # Path to the ROOT file
#    input_root_file = "/home/neo/Desktop/test_ecal_dif_5_tree.root"
    input_root_file = "/Users/akihiro/SynologyDrive/研究/WAGASCI/片山/analysis/data/physics_run_2020-02-11_22-23-18_106_ecal_dif_0_tree.root";
    # TFile object
    tfile = ROOT.TFile(input_root_file, "READ")
    # get TTree from TFile
    ttree = tfile.Get("raw")
    # create canvas

    # Print image to file
    # create TCanvas of 1280x720 pixels
    tcanvas = ROOT.TCanvas("c1", "c1", 1280, 720)
    ttree.Draw("spill_number")
    tcanvas.Print("/Users/akihiro/Desktop/test.png")

    # Loop example
    counter = 0
    for event in ttree:
        print(event.spill_number)
        # break after 10 spills to not have too long output
        counter += 1
        if counter == 10:
            break



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import os
import sys

import wagascianpy.plotting.detector
import wagascianpy.plotting.marker
import wagascianpy.plotting.parse_args
import wagascianpy.plotting.plotter
import wagascianpy.utils.utils
from wagascianpy.plotting.configuration import Configuration


def history_plotter(plotter):
    plotter.template_plotter()


if __name__ == "__main__":

    # Parse shell arguments
    args = wagascianpy.plotting.parse_args.parse_args(sys.argv[1:])

    # Edit the initial configuration
    wagascianpy.plotting.configuration.fill_configuration(args)

    if not os.path.exists(Configuration.plotter.output_path()):
        wagascianpy.utils.utils.mkdir_p(Configuration.plotter.output_path())

    markers = wagascianpy.plotting.marker.MarkerTuple(
        run=Configuration.plotter.run_markers(),
        maintenance=Configuration.plotter.maintenance_markers(),
        trouble=Configuration.plotter.trouble_markers(),
        bsd=Configuration.plotter.bsd_markers())

    output_file_format = os.path.join(Configuration.plotter.output_path(),
                                      "%s_{name}.pdf" % Configuration.plotter.output_string())

    topology_str = Configuration.plotter.topology()
    topology = wagascianpy.plotting.parse_args.parse_plotting_topology(topology_str) if topology_str else None

    if Configuration.bsddb.bsd_download_location():
        bsd_location = Configuration.bsddb.bsd_download_location()
    else:
        bsd_location = Configuration.bsddb.bsd_repository()

    if Configuration.plotter.delivered_pot():
        history_plotter(
            wagascianpy.plotting.plotter.BsdPotPlotter(
                output_file_path=output_file_format.format(name='bsd_pot_history'),
                bsd_database=Configuration.bsddb.bsd_database(),
                bsd_repository=bsd_location,
                wagasci_database=Configuration.wagascidb.wagasci_database(),
                t2krun=Configuration.global_configuration.t2krun(),
                start=Configuration.run_select.start(),
                stop=Configuration.run_select.stop(),
                markers=markers,
                save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.accumulated_pot():
        history_plotter(wagascianpy.plotting.plotter.WagasciPotPlotter(
            output_file_path=output_file_format.format(name='pot_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            only_good=Configuration.plotter.only_good(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.bsd_spill_history():
        history_plotter(wagascianpy.plotting.plotter.BsdSpillPlotter(
            output_file_path=output_file_format.format(name='bsd_spill_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.wagasci_spill_history():
        history_plotter(wagascianpy.plotting.plotter.WagasciSpillHistoryPlotter(
            output_file_path=output_file_format.format(name='wg_spill_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.wagasci_fixed_spill():
        history_plotter(wagascianpy.plotting.plotter.WagasciFixedSpillPlotter(
            output_file_path=output_file_format.format(name='wg_fixed_spill_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.wagasci_spill_number():
        history_plotter(wagascianpy.plotting.plotter.WagasciSpillNumberPlotter(
            output_file_path=output_file_format.format(name='wg_spill_number'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.temperature():
        history_plotter(wagascianpy.plotting.plotter.TemperaturePlotter(
            output_file_path=output_file_format.format(name='temperature_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.humidity():
        history_plotter(wagascianpy.plotting.plotter.HumidityPlotter(
            output_file_path=output_file_format.format(name='humidity_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.bcid_history():
        history_plotter(wagascianpy.plotting.plotter.BcidPlotter(
            output_file_path=output_file_format.format(name='bcid_history'),
            bsd_database=Configuration.bsddb.bsd_database(),
            bsd_repository=bsd_location,
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            wagasci_repository=Configuration.wagascidb.wagasci_decoded_location(),
            t2krun=Configuration.global_configuration.t2krun(),
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            markers=markers,
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.gain_history() and not Configuration.plotter.randomize_channels() > 0 \
            and Configuration.plotter.channel() == -1:
        history_plotter(wagascianpy.plotting.plotter.GainHistoryPlotterDif(
            output_file_path=output_file_format.format(name='gain_dif_history'),
            history_location=Configuration.global_configuration.history_location(),
            markers=markers,
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.dark_noise_history() and not Configuration.plotter.randomize_channels() > 0 \
            and Configuration.plotter.channel() == -1:
        history_plotter(wagascianpy.plotting.plotter.DarkNoiseHistoryPlotterDif(
            output_file_path=output_file_format.format(name='dark_noise_dif_history'),
            history_location=Configuration.global_configuration.history_location(),
            markers=markers,
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    if Configuration.plotter.threshold_history() and not Configuration.plotter.randomize_channels() > 0 \
            and Configuration.plotter.channel() == -1:
        history_plotter(wagascianpy.plotting.plotter.ThresholdHistoryPlotterDif(
            output_file_path=output_file_format.format(name='threshold_dif_history'),
            history_location=Configuration.global_configuration.history_location(),
            markers=markers,
            start=Configuration.run_select.start(),
            stop=Configuration.run_select.stop(),
            wagasci_database=Configuration.wagascidb.wagasci_database(),
            topology=topology,
            save_tfile=Configuration.run_select.save_tfile()))

    nentries = Configuration.plotter.randomize_channels()
    if nentries <= 0 and Configuration.plotter.dif() >= 0 and Configuration.plotter.chip() >= 0 and \
            Configuration.plotter.channel() >= 0:
        difs = [Configuration.plotter.dif()]
        chips = [Configuration.plotter.chip()]
        channels = [Configuration.plotter.channel()]
    elif nentries > 0:
        difs = wagascianpy.utils.utils.randomize(nentries, 0, 8)
        chips = wagascianpy.utils.utils.randomize(nentries, 0, 20)
        channels = wagascianpy.utils.utils.randomize(nentries, 0, 32)
    else:
        difs = chips = channels = []
    random_channels = zip(difs, chips, channels)
    for random_channel in random_channels:
        dif = random_channel[0]
        chip = random_channel[1]
        channel = random_channel[2]
        if wagascianpy.plotting.detector.DifIndex.is_wallmrd(dif):
            chip = chip % 3
        name = "{{}}_history_dif{}_chip{}_chan{}".format(dif, chip, channel)
        if Configuration.plotter.gain_history():
            history_plotter(wagascianpy.plotting.plotter.GainHistoryPlotterChannel(
                output_file_path=output_file_format.format(name=name.format("gain")),
                history_location=Configuration.global_configuration.history_location(),
                markers=markers,
                start=Configuration.run_select.start(),
                stop=Configuration.run_select.stop(),
                wagasci_database=Configuration.wagascidb.wagasci_database(),
                dif=dif,
                chip=chip,
                channel=channel,
                save_tfile=Configuration.run_select.save_tfile()))

        if Configuration.plotter.dark_noise_history():
            history_plotter(wagascianpy.plotting.plotter.DarkNoiseHistoryPlotterChannel(
                output_file_path=output_file_format.format(name=name.format("dark_noise")),
                history_location=Configuration.global_configuration.history_location(),
                markers=markers,
                start=Configuration.run_select.start(),
                stop=Configuration.run_select.stop(),
                wagasci_database=Configuration.wagascidb.wagasci_database(),
                dif=dif,
                chip=chip,
                channel=channel,
                save_tfile=Configuration.run_select.save_tfile()))

        if Configuration.plotter.threshold_history():
            history_plotter(wagascianpy.plotting.plotter.ThresholdHistoryPlotterChannel(
                output_file_path=output_file_format.format(name=name.format("hit_threshold")),
                history_location=Configuration.global_configuration.history_location(),
                markers=markers,
                start=Configuration.run_select.start(),
                stop=Configuration.run_select.stop(),
                wagasci_database=Configuration.wagascidb.wagasci_database(),
                dif=dif,
                chip=chip,
                channel=channel,
                save_tfile=Configuration.run_select.save_tfile()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

"""WAGASCI simple run database"""
import json
import operator
import os
import re
import xml.etree.ElementTree as ElementTree
from datetime import datetime
from typing import Union, List, Optional, Tuple

import wagascianpy.database.db_record
import wagascianpy.database.my_tinydb
import wagascianpy.utils.acq_config_xml
import wagascianpy.utils.topology
import wagascianpy.utils.utils

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3

###############################################################################
#                                  Constants                                  #
###############################################################################

# Detector topology
_WAGASCI_TOPOLOGY = {"0": 32, "1": 32, "2": 32, "3": 32, "4": 32, "5": 32, "6": 32, "7": 32,
                     "8": 32, "9": 32, "10": 32, "11": 32, "12": 32, "13": 32, "14": 32,
                     "15": 32, "16": 32, "17": 32, "18": 32, "19": 32}
_WALLMRD_TOPOLOGY = {"0": 32, "1": 32, "2": 32}
DETECTOR_TOPOLOGY = {'WallMRD north (DIF 0-1)': {"0": _WALLMRD_TOPOLOGY, "1": _WALLMRD_TOPOLOGY},
                     'WallMRD south (DIF 2-3)': {"2": _WALLMRD_TOPOLOGY, "3": _WALLMRD_TOPOLOGY},
                     'WAGASCI upstream (DIF 4-5)': {"4": _WAGASCI_TOPOLOGY, "5": _WAGASCI_TOPOLOGY},
                     'WAGASCI downstream (DIF 6-7)': {"6": _WAGASCI_TOPOLOGY, "7": _WAGASCI_TOPOLOGY}}

# Bad runs
_GLOBAL_BAD_RUNS = [1, 2, 14, 19, 20, 26, 55, 56, 57, 82, 83, 88, 91,
                    124, 128, 130, 131, 140, 141, 142, 150, 168, 174, 175, 176, 197, 376, 377] + \
                   list(range(44, 48 + 1)) + \
                   list(range(68, 77 + 1)) + \
                   list(range(107, 117 + 1)) + \
                   list(range(186, 190 + 1))
_BAD_RUNS_WAGASCI_UPSTREAM = list(range(22, 54 + 1)) + _GLOBAL_BAD_RUNS
_BAD_RUNS_WAGASCI_DOWNSTREAM = list(range(22, 54 + 1)) + _GLOBAL_BAD_RUNS
_BAD_RUNS_WALLMRD_SOUTH = [7] + list(range(61, 78 + 1)) + _GLOBAL_BAD_RUNS
_BAD_RUNS_WALLMRD_NORTH = [7] + _GLOBAL_BAD_RUNS


###############################################################################
#                               Helper functions                              #
###############################################################################

# DIF ID ################################################

def _dif_id(file_name):
    file_name_split = os.path.splitext(os.path.basename(file_name))[0].split('ecal_dif_')
    if len(file_name_split) != 2:
        raise ValueError("raw data file name is not valid : %s" % str(file_name))
    else:
        return file_name_split[1]


# Check records ################################################

def _check_records(records):
    """check that all records are sane"""
    for record in records:
        if not isinstance(record, WagasciRunRecord):
            raise ValueError("record is not of type WagasciRunRecord")
        if not record.is_ready():
            record.pretty_print()
            raise ValueError("record is not ready to be inserted")


###########################################################################
#                                RunRecord                                #
###########################################################################

class WagasciRunRecord(wagascianpy.database.db_record.DBRecord):
    """Run record"""

    def __init__(self, record=None):

        self.name = None
        self.run_type = None
        self.run_number = None
        self.good_run_flag = None
        self.wagasci_upstream_good_data_flag = None
        self.wagasci_downstream_good_data_flag = None
        self.wallmrd_north_good_data_flag = None
        self.wallmrd_south_good_data_flag = None
        self.start_time = None
        self.stop_time = None
        self.duration_h = None
        self.topology = None
        self.run_folder = None
        self.xml_config = None
        self.raw_files = None
        super(WagasciRunRecord, self).__init__(record)

    def set_bad_run(self):
        """Set all fields other than name, run_number, run_type to default values for
        bad run. The name, run_number, run_type fields must be manually set.
        """
        self.good_run_flag = False
        self.wagasci_upstream_good_data_flag = False
        self.wagasci_downstream_good_data_flag = False
        self.wallmrd_north_good_data_flag = False
        self.wallmrd_south_good_data_flag = False
        self.start_time = 0
        self.stop_time = 0
        self.duration_h = 0.
        self.topology = "undef"
        self.raw_files = {}
        self.run_folder = "undef"
        self.xml_config = "undef"

    def set_start_time(self, datetime_str):
        """Set the run start time converting datetime string to epoch timestamp
        """
        self.start_time = self.datetime2timestamp(datetime_str)

    def set_stop_time(self, datetime_str):
        """Set the run stop time converting datetime string to epoch timestamp
        """
        self.stop_time = self.datetime2timestamp(datetime_str)

    def set_duration_h(self):
        """Set the run duration in hours
        """
        if self.stop_time is None or self.start_time is None:
            raise ValueError("Set start time and stop time before duration")
        self.duration_h = float((self.stop_time - self.start_time) / 3600.0)

    def get_start_datetime(self):
        """Get start datetime
        """
        if self.start_time is None:
            return None
        return self.timestamp2datetime(self.start_time)

    def get_stop_datetime(self):
        """Get stop datetime
        """
        if self.stop_time is None:
            return None
        return self.timestamp2datetime(self.stop_time)


###########################################################################
#                            VirtualDataBase                              #
###########################################################################

def _documents2records(documents):
    # type: (List) -> List[WagasciRunRecord]
    records = []
    for document in documents:
        records.append(WagasciRunRecord(record=document))
    return records


class WagasciDataBase(wagascianpy.database.my_tinydb.MyTinyDB):
    """Virtual class to manage a database"""

    def __init__(self, db_location, repo_location=None, is_borg_repo=False, update_db=False, rebuild_db=False,
                 wagasci_libdir=None):
        # type: (str, Optional[str], bool, bool, bool, Optional[str]) -> None
        super(WagasciDataBase, self).__init__(db_location)
        self._repository = repo_location
        self._is_borg_repo = is_borg_repo
        self._wagasci_libdir = wagasci_libdir
        if rebuild_db:
            self.clear_database()
            self._is_fresh_database = True
        if update_db or rebuild_db:
            self._update_wagasci_db()

    def _update_wagasci_db(self):
        # type: (...) -> None
        self._borg = None

        self._run_type = os.path.basename(self._repository)
        self._run_record_list = []

        if ':' in self._repository and not self._is_borg_repo:
            raise NotImplementedError("Remote non borg repositories are "
                                      "not supported at the moment")

        self._borg = wagascianpy.utils.utils.which("borg")
        if self._borg is None and self._is_borg_repo:
            raise RuntimeError("borg program not found")

        # List of runs

        if self._is_borg_repo:
            borg_list = "%s list --short --log-json %s" % (self._borg, self._repository)
            run_list = wagascianpy.utils.utils.run_borg_cmd(borg_list).strip('\n').split('\n')
        else:
            run_list = sorted(wagascianpy.utils.utils.get_immediate_subdirectories(self._repository))
        # Loop over every run

        with wagascianpy.utils.utils.Cd(self._tmp_dir):
            for run_name in sorted(run_list):

                # Skip run if it is already in the database
                if self.has_record(os.path.basename(run_name)):
                    print("Skipping run %s" % run_name)
                    continue
                else:
                    print("Run %s not found in database" % run_name)

                run = WagasciRunRecord()
                run.name = run_name
                match = re.search(r'.*_([\d]+)$', run_name)
                run.run_number = None if match is None else int(match.group(1))
                run.wagasci_upstream_good_data_flag = run.run_number not in _BAD_RUNS_WAGASCI_UPSTREAM
                run.wagasci_downstream_good_data_flag = run.run_number not in _BAD_RUNS_WAGASCI_DOWNSTREAM
                run.wallmrd_north_good_data_flag = run.run_number not in _BAD_RUNS_WALLMRD_NORTH
                run.wallmrd_south_good_data_flag = run.run_number not in _BAD_RUNS_WALLMRD_SOUTH
                run.good_run_flag = any([run.wagasci_upstream_good_data_flag, run.wagasci_downstream_good_data_flag,
                                         run.wallmrd_north_good_data_flag, run.wallmrd_south_good_data_flag])
                run.run_type = self._run_type
                run.raw_files = {}

                try:
                    file_list = []
                    if self._is_borg_repo:
                        borg_list = "%s list --short --log-json %s::%s" \
                                    % (self._borg, self._repository, run_name)
                        file_list = wagascianpy.utils.utils.run_borg_cmd(borg_list).strip('\n').split('\n')
                        run.run_folder = "not found"
                        if file_list:
                            random_file = file_list[0]
                            if run_name in random_file:
                                run.run_folder = random_file.split(run_name)[0].strip('/')
                                run.run_folder = '/' + run.run_folder
                    else:
                        run_path = os.path.join(self._repository, run_name)
                        run.run_folder = run_path
                        for root, dirs, files in os.walk(run_path, followlinks=False):
                            for filename in files:
                                file_list.append(os.path.join(root, filename))

                    for file_path in file_list:
                        if '.raw' in os.path.basename(file_path):
                            dif_id = _dif_id(file_path)
                            run.raw_files[dif_id] = file_path
                        elif '.log' in os.path.basename(file_path):
                            if self._is_borg_repo:
                                borg_extract = "%s extract --log-json %s::%s %s" \
                                               % (self._borg, self._repository, run_name, file_path)
                                wagascianpy.utils.utils.run_borg_cmd(borg_extract)
                                log_file = os.path.join(self._tmp_dir, file_path)
                            else:
                                log_file = file_path
                            log = ElementTree.parse(log_file).getroot()
                            for param in log.findall("acq/param"):
                                name = param.get('name')
                                if name == "start_time":
                                    run.set_start_time(param.text)
                                if name == "stop_time":
                                    run.set_stop_time(param.text)
                            run.set_duration_h()
                        elif '.xml' in os.path.basename(file_path) and run.xml_config is None:
                            if self._is_borg_repo:
                                borg_extract = "%s extract --log-json %s::%s %s" \
                                               % (self._borg, self._repository, run_name, file_path)
                                wagascianpy.utils.utils.run_borg_cmd(borg_extract)
                                xml = os.path.join(self._tmp_dir, file_path)
                            else:
                                xml = file_path
                            try:
                                dif_topology = wagascianpy.utils.topology.get_topology_from_xml(
                                    acq_config_xml=xml, wagasci_libdir=self._wagasci_libdir)
                                run.topology = json.dumps(dif_topology)
                            except (OSError, KeyError) as error:
                                print("Error while getting topology : %s" % error)
                                print("Pyrame XML acquisition file : %s" % xml)
                                print("WAGASCI library directory : %s" % self._wagasci_libdir)
                                run.topology = "undef"
                            run.xml_config = file_path

                except Exception as error:
                    print('run %s : %s' % (run_name, str(error)))
                    run.set_bad_run()
                finally:
                    if not run.good_run_flag:
                        print("Bad run %s found in repository" % run.name)
                    else:
                        print("Good run %s found in repository" % run.name)
                    self._run_record_list.append(run)
            _check_records(self._run_record_list)

        self.update_database(self._run_record_list)

    def get_time_interval(self,
                          datetime_start,  # type: Union[datetime, str, IntTypes, float]
                          datetime_stop,  # type: Union[datetime, str, IntTypes, float]
                          only_good=True,  # type: bool
                          include_overlapping=True  # type: bool
                          ):
        # type: (...) -> List[WagasciRunRecord]
        timestamp_start = WagasciRunRecord.datetime2timestamp(datetime_start) if datetime_start is not None else None
        timestamp_stop = WagasciRunRecord.datetime2timestamp(datetime_stop) if datetime_stop is not None else None
        return _documents2records(super(WagasciDataBase, self).get_time_interval(
            timestamp_start=timestamp_start, timestamp_stop=timestamp_stop,
            only_good=only_good, include_overlapping=include_overlapping))

    def get_run_interval(self, run_number_start, run_number_stop=None, only_good=True):
        # type: (int, int, bool) -> List[WagasciRunRecord]
        return _documents2records(super(WagasciDataBase, self).get_run_interval(
            run_number_start=run_number_start, run_number_stop=run_number_stop,
            only_good=only_good))

    def get_all(self, only_good=True):
        # type: (bool) -> List[WagasciRunRecord]
        return _documents2records(super(WagasciDataBase, self).get_all(only_good=only_good))

    def print_run(self, name):
        # type: (str) -> None
        """ Print info about a run """
        WagasciRunRecord(self.get_record(name)[0]).pretty_print()


###########################################################################
#                   Run number to time interval converter                 #
###########################################################################


def run_to_interval(start,  # type: Union[IntTypes, datetime]
                    stop=None,  # type: Optional[Union[IntTypes, datetime]]
                    database=None  # type: Optional[str]
                    ):
    # type: (...) -> Tuple[datetime, datetime]
    if isinstance(start, IntTypes) and not database:
        raise ValueError("Please provide a valid WAGASCI database")
    if isinstance(start, datetime) and isinstance(stop, datetime):
        start_time = start
        stop_time = stop
    elif isinstance(start, datetime) and not stop:
        stop = datetime.now()
        start_time = start
        stop_time = stop
    else:
        assert start is not None and isinstance(start, IntTypes), \
            "Type of start argument {} is not supported".format(type(start).__name__)
        assert stop is not None and isinstance(stop, IntTypes), \
            "Type of stop argument {} is not supported".format(type(start).__name__)
        with wagascianpy.database.wagascidb.WagasciDataBase(db_location=database) as db:
            if not stop:
                stop = db.get_last_run_number(only_good=False)
            records = db.get_run_interval(run_number_start=start, run_number_stop=stop)
            sorted_records = sorted(records, key=operator.itemgetter("run_number"))
            if sorted_records:
                start_time = wagascianpy.database.db_record.DBRecord.timestamp2datetime(
                    sorted_records[0]["start_time"])
                stop_time = wagascianpy.database.db_record.DBRecord.timestamp2datetime(
                    sorted_records[-1]["stop_time"])
            else:
                start_time = None
                stop_time = None
    if start_time is None or stop_time is None:
        raise StopIteration("Failed to get runs from {} to {} from database {}".format(start, stop, database))
    if stop_time <= start_time:
        raise ValueError("Stop time (%s) comes before start time (%s)" % (stop_time, start_time))
    start_time = wagascianpy.database.wagascidb.WagasciRunRecord.add_timezone(start_time)
    stop_time = wagascianpy.database.wagascidb.WagasciRunRecord.add_timezone(stop_time)
    return start_time, stop_time
